import datetime
from collections.abc import Iterator, Sequence
from typing import Optional, Union, cast

from dagster import _check as check
from dagster._annotations import beta
from dagster._core.definitions.asset_check_factories.utils import (
    FRESH_UNTIL_METADATA_KEY,
    ensure_no_duplicate_asset_checks,
    seconds_in_words,
)
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.decorators import sensor
from dagster._core.definitions.run_request import RunRequest, SkipReason
from dagster._core.definitions.sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorEvaluationContext,
)
from dagster._core.storage.asset_check_execution_record import AssetCheckExecutionRecordStatus
from dagster._core.storage.tags import SENSOR_NAME_TAG
from dagster._time import get_current_datetime, get_current_timestamp

DEFAULT_FRESHNESS_SENSOR_NAME = "freshness_checks_sensor"
MAXIMUM_RUNTIME_SECONDS = 35  # Due to GRPC communications, only allow this sensor to run for 40 seconds before pausing iteration and resuming in the next run. Leave a bit of time for run requests to be processed.
FRESHNESS_SENSOR_DESCRIPTION = """
    This sensor launches execution of freshness checks for the provided assets. The sensor will
    only launch a new execution of a freshness check if the check previously passed, but enough
    time has passed that the check could be overdue again. Once a check has failed, the sensor
    will not launch a new execution until the asset has been updated (which should automatically
    execute the check).
    """


@beta
def build_sensor_for_freshness_checks(
    *,
    freshness_checks: Sequence[AssetChecksDefinition],
    minimum_interval_seconds: Optional[int] = None,
    name: str = DEFAULT_FRESHNESS_SENSOR_NAME,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
) -> SensorDefinition:
    """Builds a sensor which kicks off evaluation of freshness checks.

    This sensor will kick off an execution of a check in the following cases:
    - The check has never been executed before.
    - The check has been executed before, and the previous result was a success, but it is again
    possible for the check to be overdue based on the `dagster/fresh_until_timestamp` metadata
    on the check result.

    Note that we will not execute if:
    - The freshness check has been executed before, and the previous result was a failure. This is
    because whichever run materializes/observes the run to bring the check back to a passing
    state will end up also running the check anyway, so until that run occurs, there's no point
    in evaluating the check.
    - The freshness check has been executed before, and the previous result was a success, but it is
    not possible for the check to be overdue based on the `dagster/fresh_until_timestamp`
    metadata on the check result. Since the check cannot be overdue, we know the check
    result would not change with an additional execution.

    Args:
        freshness_checks (Sequence[AssetChecksDefinition]): The freshness checks to evaluate.
        minimum_interval_seconds (Optional[int]): The duration in seconds between evaluations of the sensor.
        name (Optional[str]): The name of the sensor. Defaults to "freshness_check_sensor", but a
            name may need to be provided in case of multiple calls of this function.
        default_status (Optional[DefaultSensorStatus]): The default status of the sensor. Defaults
            to stopped.

    Returns:
        SensorDefinition: The sensor that kicks off freshness evaluations.
    """
    freshness_checks = check.sequence_param(freshness_checks, "freshness_checks")
    ensure_no_duplicate_asset_checks(freshness_checks)
    check.invariant(
        check.int_param(minimum_interval_seconds, "minimum_interval_seconds") > 0
        if minimum_interval_seconds
        else True,
        "Interval must be a positive integer.",
    )
    check.str_param(name, "name")

    @sensor(
        name=name,
        minimum_interval_seconds=minimum_interval_seconds,
        asset_selection=AssetSelection.checks(*freshness_checks),
        default_status=default_status,
        description=FRESHNESS_SENSOR_DESCRIPTION,
    )
    def the_sensor(context: SensorEvaluationContext) -> Optional[Union[RunRequest, SkipReason]]:
        left_off_asset_check_key = (
            AssetCheckKey.from_user_string(context.cursor) if context.cursor else None
        )
        start_time = get_current_datetime()
        checks_to_evaluate = []
        checks_iter = freshness_checks_get_evaluations_iter(
            context=context,
            start_time=start_time,
            left_off_asset_check_key=left_off_asset_check_key,
            freshness_checks=freshness_checks,
        )
        # We evaluate checks using an iterator which yields back control to the main loop every
        # iteration; this allows us to pause the sensor if it runs into the maximum runtime.
        check_key, should_evaluate = next(checks_iter, (None, False))
        while (
            get_current_timestamp() - start_time.timestamp() < MAXIMUM_RUNTIME_SECONDS and check_key
        ):
            if should_evaluate:
                checks_to_evaluate.append(check_key)
            check_key, should_evaluate = next(checks_iter, (None, False))
        new_cursor = check_key.to_user_string() if check_key else None
        context.update_cursor(new_cursor)
        if checks_to_evaluate:
            return RunRequest(asset_check_keys=checks_to_evaluate)
        else:
            return SkipReason(
                "No freshness checks need to be evaluated at this time, since all checks are either currently evaluating, have failed, or are not yet overdue."
            )

    return the_sensor


def ordered_iterator_freshness_checks_starting_with_key(
    left_off_asset_check_key: Optional[AssetCheckKey],
    freshness_checks: Sequence[AssetChecksDefinition],
) -> Iterator[AssetCheckKey]:
    asset_check_keys_sorted = sorted(
        [
            asset_check_spec.key
            for asset_check in freshness_checks
            for asset_check_spec in asset_check.check_specs
        ],
        key=lambda key: key.to_user_string(),
    )
    # Offset based on the left off asset check key, but then iterate back through the beginning afterwards
    if left_off_asset_check_key:
        left_off_idx = asset_check_keys_sorted.index(left_off_asset_check_key)
        yield from asset_check_keys_sorted[left_off_idx + 1 :]
        yield from asset_check_keys_sorted[: left_off_idx + 1]
    else:
        yield from asset_check_keys_sorted


def freshness_checks_get_evaluations_iter(
    context: SensorEvaluationContext,
    start_time: datetime.datetime,
    left_off_asset_check_key: Optional[AssetCheckKey],
    freshness_checks: Sequence[AssetChecksDefinition],
) -> Iterator[tuple[AssetCheckKey, bool]]:
    """Yields the set of freshness check keys to evaluate."""
    for check_key in ordered_iterator_freshness_checks_starting_with_key(
        left_off_asset_check_key, freshness_checks
    ):
        summary_record = context.instance.event_log_storage.get_asset_check_summary_records(
            [check_key]
        )[check_key]
        # Case 1: We have never run the check before. We should run it.
        if summary_record.last_check_execution_record is None:
            context.log.info(
                f"Freshness check {check_key.to_user_string()} has never been executed before. "
                "Adding to run request."
            )
            yield check_key, True
            continue

        # Case 2: The check is currently evaluating and has never previously evaluated. We shouldn't kick off another evaluation until it's done.
        if summary_record.last_completed_check_execution_record is None:
            check.invariant(
                summary_record.last_check_execution_record.status
                == AssetCheckExecutionRecordStatus.PLANNED,
                f"Unexpected status for check {check_key.to_user_string()}. The summary record indicates that the check has never completed, but the last check execution record is in terminal state {summary_record.last_check_execution_record.status}. This is likely a framework error, please report this to the Dagster maintainers.",
            )

            context.log.info(
                f"Freshness check on asset {check_key.asset_key.to_user_string()} is currently evaluating for the first time. Skipping..."
            )
            yield check_key, False
            continue

        latest_completed_record = check.not_none(
            summary_record.last_completed_check_execution_record
        )
        latest_record_any_status = check.not_none(summary_record.last_check_execution_record)
        latest_completed_evaluation = check.not_none(
            check.not_none(latest_completed_record.event).asset_check_evaluation
        )
        # Case 3: The check is currently evaluating and the previous evaluation passed.
        # - If the check is overdue and the current evaluation was not kicked off by the sensor, we should kick off another evaluation.
        # - If the check previously failed, we'll wait for the current evaluation to complete to avoid over-evaluation.
        if (
            latest_record_any_status.status == AssetCheckExecutionRecordStatus.PLANNED
            and latest_completed_evaluation.passed
        ):
            run_record = context.instance.get_run_record_by_id(latest_record_any_status.run_id)
            # Previous run was kicked off by the sensor, wait for it to complete before kicking off another run.
            if (
                run_record
                and run_record.dagster_run.tags.get(SENSOR_NAME_TAG) == context.sensor_name
            ):
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} is currently evaluating, and was kicked off by this sensor. Skipping..."
                )
                yield check_key, False
                continue
            # Previous run was not kicked off by the sensor, check if it's overdue.
            next_deadline = cast(
                float, latest_completed_evaluation.metadata[FRESH_UNTIL_METADATA_KEY].value
            )
            if next_deadline < start_time.timestamp():
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} is currently evaluating, but "
                    "enough time has passed that it can be overdue again. Adding to run request."
                )
                yield check_key, True
                continue
            else:
                how_long_until_next_deadline = next_deadline - start_time.timestamp()
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} is currently evaluating, but "
                    f"cannot be overdue again until {seconds_in_words(how_long_until_next_deadline)} from now. Skipping..."
                )
                yield check_key, False
                continue
        # Case 4: The previous completed evaluation failed. We should kick off another only if the asset has been updated since it's last evaluation.
        elif not latest_completed_evaluation.passed:
            latest_materialization = context.instance.get_latest_materialization_event(
                check_key.asset_key
            )
            # If the asset has been updated since the last evaluation, we should re-evaluate the check.
            if (
                latest_materialization
                and latest_materialization.timestamp
                > check.not_none(latest_completed_record.event).timestamp
            ):
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} previously failed, but "
                    "the asset has been updated since the last evaluation. Adding to run request."
                )
                yield check_key, True
                continue
            else:
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} previously failed, but "
                    "the asset has not been updated since the last evaluation. Skipping..."
                )
                yield check_key, False
                continue
        # Case 5: The previous evaluation passed and there is no in progress evaluation. We should kick off another evaluation only if the check is overdue.
        else:
            next_deadline = cast(
                float, latest_completed_evaluation.metadata[FRESH_UNTIL_METADATA_KEY].value
            )
            if next_deadline < start_time.timestamp():
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} previously passed, but "
                    "enough time has passed that it can be overdue again. Adding to run request."
                )
                yield check_key, True
                continue
            else:
                how_long_until_next_deadline = next_deadline - start_time.timestamp()
                context.log.info(
                    f"Freshness check {check_key.to_user_string()} previously passed, but "
                    f"cannot be overdue again until {seconds_in_words(how_long_until_next_deadline)} from now. Skipping..."
                )
                yield check_key, False
                continue



import os
import sys
from collections.abc import Sequence
from typing import NamedTuple, Optional, Union

import click
import requests

from dagster._generate import download_example_from_github, generate_project, generate_repository
from dagster._generate.download import AVAILABLE_EXAMPLES
from dagster.version import __version__ as dagster_version


@click.group(name="project")
def project_cli() -> None:
    """Commands for bootstrapping new Dagster projects and code locations."""


# Keywords to flag in package names. When a project name contains one of these keywords, we check
# if a conflicting PyPI package exists.
FLAGGED_PACKAGE_KEYWORDS = ["dagster", "dbt"]
VALID_EXCLUDES: list[str] = ["readme.md", "setup", "tests"]  # all lower case

scaffold_repository_command_help_text = (
    "(DEPRECATED; Use `dagster project scaffold --excludes README.md` instead) "
    "Create a folder structure with a single Dagster repository, in the current directory. "
    "This CLI helps you to scaffold a new Dagster repository within a folder structure that "
    "includes multiple Dagster repositories"
)

scaffold_code_location_command_help_text = (
    "(DEPRECATED; Use `dagster project scaffold --excludes README.md` instead) "
    "Create a folder structure with a single Dagster code location, in the current directory. "
    "This CLI helps you to scaffold a new Dagster code location within a folder structure that "
    "includes multiple Dagster code locations."
)

scaffold_command_help_text = (
    "Create a folder structure with a single Dagster code location and other files such as "
    "pyproject.toml. This CLI enables you to quickly start building a new Dagster project with "
    "everything set up."
)

from_example_command_help_text = (
    "Download one of the official Dagster examples to the current directory. "
    "This CLI enables you to quickly bootstrap your project with an officially maintained example."
)

list_examples_command_help_text = "List the examples that available to bootstrap with."


class PackageConflictCheckResult(NamedTuple):
    request_error_msg: Optional[str]
    conflict_exists: bool = False


def check_if_pypi_package_conflict_exists(project_name: str) -> PackageConflictCheckResult:
    """Checks if the project name contains any flagged keywords. If so, raises a warning if a PyPI
    package with the same name exists. This is to prevent import errors from occurring due to a
    project name that conflicts with an imported package.

    Raises an error regardless of hyphen or underscore (i.e. dagster_dbt vs dagster-dbt). Both
    are invalid and cause import errors.
    """
    if any(keyword in project_name for keyword in FLAGGED_PACKAGE_KEYWORDS):
        try:
            res = requests.get(f"https://pypi.org/pypi/{project_name}/json")
            if res.status_code == 200:
                return PackageConflictCheckResult(request_error_msg=None, conflict_exists=True)
        except Exception as e:
            return PackageConflictCheckResult(request_error_msg=str(e))

    return PackageConflictCheckResult(request_error_msg=None, conflict_exists=False)


def _check_and_error_on_package_conflicts(project_name: str) -> None:
    package_check_result: PackageConflictCheckResult = check_if_pypi_package_conflict_exists(
        project_name
    )
    if package_check_result.request_error_msg:
        click.echo(
            click.style(
                "An error occurred while checking for package conflicts:"
                f" {package_check_result.request_error_msg}. \n\nConflicting package names will"
                " cause import errors in your project if the existing PyPI package is included"
                " as a dependency in your scaffolded project. If desired, this check can be"
                " skipped by adding the `--ignore-package-conflict` flag.",
                fg="red",
            )
        )
        sys.exit(1)

    if package_check_result.conflict_exists:
        click.echo(
            click.style(
                f"The project '{project_name}' conflicts with an existing PyPI package."
                " Conflicting package names will cause import errors in your project if the"
                " existing PyPI package is included as a dependency in your scaffolded"
                " project. Please choose another name, or add the `--ignore-package-conflict`"
                " flag to bypass this check.",
                fg="yellow",
            )
        )
        sys.exit(1)


@project_cli.command(
    name="scaffold-repository",
    short_help=scaffold_repository_command_help_text,
    help=scaffold_repository_command_help_text,
)
@click.option(
    "--name",
    required=True,
    type=click.STRING,
    help="Name of the new Dagster repository",
)
def scaffold_repository_command(name: str) -> None:
    dir_abspath = os.path.abspath(name)
    if os.path.isdir(dir_abspath) and os.path.exists(dir_abspath):
        click.echo(
            click.style(f"The directory {dir_abspath} already exists. ", fg="red")
            + "\nPlease delete the contents of this path or choose another location."
        )
        sys.exit(1)

    click.echo(
        click.style(
            "WARNING: command is deprecated. Use `dagster project scaffold --excludes readme` instead.",
            fg="yellow",
        )
    )
    generate_repository(dir_abspath)
    click.echo(_styled_success_statement(name, dir_abspath))


@project_cli.command(
    name="scaffold-code-location",
    short_help=scaffold_code_location_command_help_text,
    help=scaffold_code_location_command_help_text,
)
@click.option(
    "--name",
    required=True,
    type=click.STRING,
    help="Name of the new Dagster code location",
)
@click.pass_context
def scaffold_code_location_command(context, name: str):
    click.echo(
        click.style(
            "WARNING: command is deprecated. Use `dagster project scaffold --excludes readme` instead.",
            fg="yellow",
        )
    )
    context.invoke(scaffold_command, name=name, excludes=["README.md"])


@project_cli.command(
    name="scaffold",
    short_help=scaffold_command_help_text,
    help=scaffold_command_help_text,
)
@click.option(
    "--name",
    required=True,
    type=click.STRING,
    help="Name of the new Dagster project",
)
@click.option(
    "--excludes",
    multiple=True,
    type=click.STRING,
    default=[],
    help=f"Exclude case-insensitive file patterns from the project template. Valid patterns: {VALID_EXCLUDES}",
)
@click.option(
    "--ignore-package-conflict",
    is_flag=True,
    default=False,
    help="Controls whether the project name can conflict with an existing PyPI package.",
)
def scaffold_command(
    name: str,
    excludes: Optional[Union[list[str], tuple]] = None,
    ignore_package_conflict: bool = False,
) -> None:
    dir_abspath = os.path.abspath(name)
    if os.path.isdir(dir_abspath) and os.path.exists(dir_abspath):
        click.echo(
            click.style(f"The directory {dir_abspath} already exists. ", fg="red")
            + "\nPlease delete the contents of this path or choose another location."
        )
        sys.exit(1)

    if isinstance(excludes, tuple):
        excludes = list(excludes)
    excludes = [] if not excludes else [item for item in excludes]

    invalid_excludes = [item for item in excludes if item.lower() not in VALID_EXCLUDES]
    if invalid_excludes:
        click.echo(
            click.style(
                f"The following strings are invalid options for the excludes tag: {invalid_excludes}",
                fg="red",
            )
            + f"Choose from {VALID_EXCLUDES}. Case-insensitive. "
        )
        sys.exit(1)
    excludes = [item.lower() for item in excludes if item in VALID_EXCLUDES]

    if not ignore_package_conflict:
        _check_and_error_on_package_conflicts(name)

    generate_project(dir_abspath, excludes=excludes)
    click.echo(_styled_success_statement(name, dir_abspath))


@project_cli.command(
    name="from-example",
    short_help=from_example_command_help_text,
    help=from_example_command_help_text,
)
@click.option(
    "--name",
    type=click.STRING,
    help="Name of the new Dagster project",
)
@click.option(
    "--example",
    required=True,
    type=click.STRING,
    help=(
        "Name of the example to bootstrap with. You can use an example name from the official "
        "examples in Dagster repo: https://github.com/dagster-io/dagster/tree/master/examples. "
        "You can also find the available examples via `dagster project list-examples`."
    ),
)
@click.option(
    "--version",
    type=click.STRING,
    help="Which version of the example to download, defaults to same version as installed dagster.",
    default=dagster_version,
    show_default=True,
)
def from_example_command(name: Optional[str], example: str, version: str) -> None:
    name = name or example
    dir_abspath = os.path.abspath(name) + "/"
    if os.path.isdir(dir_abspath) and os.path.exists(dir_abspath):
        click.echo(
            click.style(f"The directory {dir_abspath} already exists. ", fg="red")
            + "\nPlease delete the contents of this path or choose another location."
        )
        sys.exit(1)
    else:
        os.makedirs(dir_abspath, exist_ok=True)

    download_example_from_github(dir_abspath, example, version)

    click.echo(_styled_success_statement(name, dir_abspath))


@project_cli.command(
    name="list-examples",
    short_help=list_examples_command_help_text,
    help=list_examples_command_help_text,
)
def from_example_list_command() -> None:
    click.echo("Examples available in `dagster project from-example`:")

    click.echo(_styled_list_examples_prints(AVAILABLE_EXAMPLES))


def _styled_list_examples_prints(examples: Sequence[str]) -> str:
    return "\n".join([f"* {name}" for name in examples])


def _styled_success_statement(name: str, path: str) -> str:
    return (
        click.style("Success!", fg="green")
        + " Created "
        + click.style(name, fg="blue")
        + " at "
        + click.style(path, fg="blue")
        + "."
    )



from collections.abc import Sequence
from datetime import datetime, timedelta
from typing import NamedTuple, Optional, cast

import dagster._check as check
from dagster._annotations import PublicAttr, beta_param
from dagster._core.definitions.partition import (
    AllPartitionsSubset,
    PartitionsDefinition,
    PartitionsSubset,
)
from dagster._core.definitions.partition_mapping import PartitionMapping, UpstreamPartitionsResult
from dagster._core.definitions.time_window_partitions import (
    TimeWindow,
    TimeWindowPartitionsDefinition,
    TimeWindowPartitionsSubset,
)
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.instance import DynamicPartitionsStore
from dagster._serdes import whitelist_for_serdes
from dagster._time import add_absolute_time


@whitelist_for_serdes
@beta_param(param="allow_nonexistent_upstream_partitions")
class TimeWindowPartitionMapping(
    PartitionMapping,
    NamedTuple(
        "_TimeWindowPartitionMapping",
        [
            ("start_offset", PublicAttr[int]),
            ("end_offset", PublicAttr[int]),
            ("allow_nonexistent_upstream_partitions", PublicAttr[bool]),
        ],
    ),
):
    """The default mapping between two TimeWindowPartitionsDefinitions.

    A partition in the downstream partitions definition is mapped to all partitions in the upstream
    asset whose time windows overlap it.

    This means that, if the upstream and downstream partitions definitions share the same time
    period, then this mapping is essentially the identity partition mapping - plus conversion of
    datetime formats.

    If the upstream time period is coarser than the downstream time period, then each partition in
    the downstream asset will map to a single (larger) upstream partition. E.g. if the downstream is
    hourly and the upstream is daily, then each hourly partition in the downstream will map to the
    daily partition in the upstream that contains that hour.

    If the upstream time period is finer than the downstream time period, then each partition in the
    downstream asset will map to multiple upstream partitions. E.g. if the downstream is daily and
    the upstream is hourly, then each daily partition in the downstream asset will map to the 24
    hourly partitions in the upstream that occur on that day.

    Args:
        start_offset (int): If not 0, then the starts of the upstream windows are shifted by this
            offset relative to the starts of the downstream windows. For example, if start_offset=-1
            and end_offset=0, then the downstream partition "2022-07-04" would map to the upstream
            partitions "2022-07-03" and "2022-07-04". If the upstream and downstream
            PartitionsDefinitions are different, then the offset is in the units of the downstream.
            Defaults to 0.
        end_offset (int): If not 0, then the ends of the upstream windows are shifted by this
            offset relative to the ends of the downstream windows. For example, if start_offset=0
            and end_offset=1, then the downstream partition "2022-07-04" would map to the upstream
            partitions "2022-07-04" and "2022-07-05". If the upstream and downstream
            PartitionsDefinitions are different, then the offset is in the units of the downstream.
            Defaults to 0.
        allow_nonexistent_upstream_partitions (bool): Defaults to false. If true, does not
            raise an error when mapped upstream partitions fall outside the start-end time window of the
            partitions def. For example, if the upstream partitions def starts on "2023-01-01" but
            the downstream starts on "2022-01-01", setting this bool to true would return no
            partition keys when get_upstream_partitions_for_partitions is called with "2022-06-01".
            When set to false, would raise an error.

    Examples:
        .. code-block:: python

            from dagster import DailyPartitionsDefinition, TimeWindowPartitionMapping, AssetIn, asset

            partitions_def = DailyPartitionsDefinition(start_date="2020-01-01")

            @asset(partitions_def=partitions_def)
            def asset1():
                ...

            @asset(
                partitions_def=partitions_def,
                ins={
                    "asset1": AssetIn(
                        partition_mapping=TimeWindowPartitionMapping(start_offset=-1)
                    )
                }
            )
            def asset2(asset1):
                ...
    """

    def __new__(
        cls,
        start_offset: int = 0,
        end_offset: int = 0,
        allow_nonexistent_upstream_partitions: bool = False,
    ):
        return super().__new__(
            cls,
            start_offset=check.int_param(start_offset, "start_offset"),
            end_offset=check.int_param(end_offset, "end_offset"),
            allow_nonexistent_upstream_partitions=check.bool_param(
                allow_nonexistent_upstream_partitions,
                "allow_nonexistent_upstream_partitions",
            ),
        )

    @property
    def description(self) -> str:
        description_str = (
            "Maps a downstream partition to any upstream partition with an overlapping time window."
        )

        if self.start_offset != 0 or self.end_offset != 0:
            description_str += (
                f" The start and end of the upstream time window is offsetted by "
                f"{self.start_offset} and {self.end_offset} partitions respectively."
            )

        return description_str

    def _validated_input_partitions_subset(
        self, param_name: str, subset: Optional[PartitionsSubset]
    ) -> TimeWindowPartitionsSubset:
        if isinstance(subset, AllPartitionsSubset):
            return TimeWindowPartitionsSubset.from_all_partitions_subset(subset)
        else:
            return check.inst_param(
                cast(TimeWindowPartitionsSubset, subset),
                param_name,
                TimeWindowPartitionsSubset,
            )

    def _validated_input_partitions_def(
        self, param_name: str, partitions_def: Optional[PartitionsDefinition]
    ) -> TimeWindowPartitionsDefinition:
        return check.inst_param(
            cast(TimeWindowPartitionsDefinition, partitions_def),
            param_name,
            TimeWindowPartitionsDefinition,
        )

    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: Optional[PartitionsSubset],
        downstream_partitions_def: Optional[PartitionsDefinition],
        upstream_partitions_def: PartitionsDefinition,
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> UpstreamPartitionsResult:
        return self._map_partitions(
            from_partitions_def=self._validated_input_partitions_def(
                "downstream_partitions_def", downstream_partitions_def
            ),
            to_partitions_def=self._validated_input_partitions_def(
                "upstream_partitions_def", upstream_partitions_def
            ),
            from_partitions_subset=self._validated_input_partitions_subset(
                "downstream_partitions_subset", downstream_partitions_subset
            ),
            start_offset=self.start_offset,
            end_offset=self.end_offset,
            current_time=current_time,
            mapping_downstream_to_upstream=True,
        )

    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: Optional[PartitionsDefinition],
        current_time: Optional[datetime] = None,
        dynamic_partitions_store: Optional[DynamicPartitionsStore] = None,
    ) -> PartitionsSubset:
        """Returns the partitions in the downstream asset that map to the given upstream partitions.

        Filters for partitions that exist at the given current_time, fetching the current time
        if not provided.
        """
        return self._map_partitions(
            from_partitions_def=self._validated_input_partitions_def(
                "upstream_partitions_def", upstream_partitions_def
            ),
            to_partitions_def=self._validated_input_partitions_def(
                "downstream_partitions_def", downstream_partitions_def
            ),
            from_partitions_subset=self._validated_input_partitions_subset(
                "upstream_partitions_subset", upstream_partitions_subset
            ),
            end_offset=-self.start_offset,
            start_offset=-self.end_offset,
            current_time=current_time,
            mapping_downstream_to_upstream=False,
        ).partitions_subset

    def _merge_time_windows(self, time_windows: Sequence[TimeWindow]) -> Sequence[TimeWindow]:
        """Takes a list of potentially-overlapping TimeWindows and merges any overlapping windows."""
        if not time_windows:
            return []

        sorted_time_windows = sorted(time_windows, key=lambda tw: tw.start.timestamp())
        merged_time_windows: list[TimeWindow] = [sorted_time_windows[0]]

        for window in sorted_time_windows[1:]:
            last_window = merged_time_windows[-1]
            # window starts before the end of the previous one
            if window.start.timestamp() <= last_window.end.timestamp():
                merged_time_windows[-1] = TimeWindow(
                    last_window.start, max(last_window.end, window.end, key=lambda d: d.timestamp())
                )
            else:
                merged_time_windows.append(window)

        return merged_time_windows

    def _map_partitions(
        self,
        from_partitions_def: TimeWindowPartitionsDefinition,
        to_partitions_def: TimeWindowPartitionsDefinition,
        from_partitions_subset: TimeWindowPartitionsSubset,
        start_offset: int,
        end_offset: int,
        current_time: Optional[datetime],
        mapping_downstream_to_upstream: bool,
    ) -> UpstreamPartitionsResult:
        """Maps the partitions in from_partitions_subset to partitions in to_partitions_def.

        If partitions in from_partitions_subset represent time windows that do not exist in
        to_partitions_def, raises an error if raise_error_on_invalid_mapped_partition is True.
        Otherwise, filters out the partitions that do not exist in to_partitions_def and returns
        the filtered subset, also returning a bool indicating whether there were mapped time windows
        that did not exist in to_partitions_def.

        Args:
            mapping_downstream_to_upstream (bool): True if from_partitions_def is the downstream
                partitions def and to_partitions_def is the upstream partitions def.
        """
        if not isinstance(from_partitions_subset, TimeWindowPartitionsSubset):
            check.failed("from_partitions_subset must be a TimeWindowPartitionsSubset")

        if not isinstance(from_partitions_def, TimeWindowPartitionsDefinition):
            check.failed("from_partitions_def must be a TimeWindowPartitionsDefinition")

        if not isinstance(to_partitions_def, TimeWindowPartitionsDefinition):
            check.failed("to_partitions_def must be a TimeWindowPartitionsDefinition")

        if to_partitions_def.timezone != from_partitions_def.timezone:
            raise DagsterInvalidDefinitionError(
                f"Timezones {to_partitions_def.timezone} and {from_partitions_def.timezone} don't match"
            )

        result = self._do_cheap_partition_mapping_if_possible(
            from_partitions_def=from_partitions_def,
            to_partitions_def=to_partitions_def,
            from_partitions_subset=from_partitions_subset,
            start_offset=start_offset,
            end_offset=end_offset,
            current_time=current_time,
        )
        if result is not None:
            return result

        first_window = to_partitions_def.get_first_partition_window(current_time=current_time)
        last_window = to_partitions_def.get_last_partition_window(current_time=current_time)
        full_window = (
            TimeWindow(first_window.start, last_window.end)
            if first_window is not None and last_window is not None
            else None
        )

        time_windows = []
        for from_partition_time_window in from_partitions_subset.included_time_windows:
            from_start_dt, from_end_dt = (
                from_partition_time_window.start,
                from_partition_time_window.end,
            )

            if mapping_downstream_to_upstream:
                offsetted_from_start_dt = _offsetted_datetime_with_bounds(
                    from_partitions_def, from_start_dt, start_offset, full_window
                )
                offsetted_from_end_dt = _offsetted_datetime_with_bounds(
                    from_partitions_def, from_end_dt, end_offset, full_window
                )
            else:
                # we'll apply the offsets later, after we've found the corresponding windows
                offsetted_from_start_dt = from_start_dt
                offsetted_from_end_dt = from_end_dt

            # Align the windows to partition boundaries in the target PartitionsDefinition
            if (from_partitions_def.cron_schedule == to_partitions_def.cron_schedule) or (
                from_partitions_def.is_basic_daily and to_partitions_def.is_basic_hourly
            ):
                # If the above conditions hold true, then we're confident that the partition
                # boundaries in the PartitionsDefinition that we're mapping from match up with
                # boundaries in the PartitionsDefinition that we're mapping to. That means
                # we can just use these boundaries directly instead of finding nearby boundaries.
                to_start_dt = offsetted_from_start_dt
                to_end_dt = offsetted_from_end_dt
            else:
                # The partition boundaries that we're mapping from might land in the middle of
                # partitions that we're mapping to, so find those partitions.
                to_start_partition_key = to_partitions_def.get_partition_key_for_timestamp(
                    offsetted_from_start_dt.timestamp(), end_closed=False
                )
                to_end_partition_key = to_partitions_def.get_partition_key_for_timestamp(
                    offsetted_from_end_dt.timestamp(), end_closed=True
                )

                to_start_dt = to_partitions_def.start_time_for_partition_key(to_start_partition_key)
                to_end_dt = to_partitions_def.end_time_for_partition_key(to_end_partition_key)

            if mapping_downstream_to_upstream:
                offsetted_to_start_dt = to_start_dt
                offsetted_to_end_dt = to_end_dt
            else:
                offsetted_to_start_dt = _offsetted_datetime_with_bounds(
                    to_partitions_def, to_start_dt, start_offset, full_window
                )
                offsetted_to_end_dt = _offsetted_datetime_with_bounds(
                    to_partitions_def, to_end_dt, end_offset, full_window
                )

            if offsetted_to_start_dt.timestamp() < offsetted_to_end_dt.timestamp():
                time_windows.append(TimeWindow(offsetted_to_start_dt, offsetted_to_end_dt))

        filtered_time_windows = []
        required_but_nonexistent_subset = to_partitions_def.empty_subset()

        for time_window in time_windows:
            if (
                first_window
                and last_window
                and time_window.start.timestamp() <= last_window.start.timestamp()
                and time_window.end.timestamp() >= first_window.end.timestamp()
            ):
                window_start = max(
                    time_window.start, first_window.start, key=lambda d: d.timestamp()
                )
                window_end = min(time_window.end, last_window.end, key=lambda d: d.timestamp())
                filtered_time_windows.append(TimeWindow(window_start, window_end))

            if self.allow_nonexistent_upstream_partitions:
                # If allowed to have nonexistent upstream partitions, do not consider
                # out of range partitions to be invalid
                continue
            else:
                invalid_time_window = None
                if not (first_window and last_window) or (
                    time_window.start.timestamp() < first_window.start.timestamp()
                    and time_window.end.timestamp() > last_window.end.timestamp()
                ):
                    invalid_time_window = time_window
                elif time_window.start.timestamp() < first_window.start.timestamp():
                    invalid_time_window = TimeWindow(
                        time_window.start,
                        min(time_window.end, first_window.start, key=lambda d: d.timestamp()),
                    )
                elif time_window.end.timestamp() > last_window.end.timestamp():
                    invalid_time_window = TimeWindow(
                        max(time_window.start, last_window.end, key=lambda d: d.timestamp()),
                        time_window.end,
                    )

                if invalid_time_window:
                    required_but_nonexistent_subset = (
                        required_but_nonexistent_subset
                        | to_partitions_def.get_partition_subset_in_time_window(
                            time_window=invalid_time_window
                        )
                    )

        return UpstreamPartitionsResult(
            partitions_subset=TimeWindowPartitionsSubset(
                to_partitions_def,
                num_partitions=None,
                included_time_windows=self._merge_time_windows(filtered_time_windows),
            ),
            required_but_nonexistent_subset=required_but_nonexistent_subset,
        )

    def _do_cheap_partition_mapping_if_possible(
        self,
        from_partitions_def: TimeWindowPartitionsDefinition,
        to_partitions_def: TimeWindowPartitionsDefinition,
        from_partitions_subset: TimeWindowPartitionsSubset,
        start_offset: int,
        end_offset: int,
        current_time: Optional[datetime],
    ) -> Optional[UpstreamPartitionsResult]:
        """The main partition-mapping logic relies heavily on expensive cron iteration operations.

        This method covers a set of easy cases where these operations aren't required. It returns
        None if the mapping doesn't fit into any of these cases.
        """
        if from_partitions_subset.is_empty:
            return UpstreamPartitionsResult(
                partitions_subset=to_partitions_def.empty_subset(),
                required_but_nonexistent_subset=to_partitions_def.empty_subset(),
            )

        if start_offset != 0 or end_offset != 0:
            return None

        # Same PartitionsDefinitions
        if from_partitions_def == to_partitions_def:
            return UpstreamPartitionsResult(
                partitions_subset=from_partitions_subset,
                required_but_nonexistent_subset=to_partitions_def.empty_subset(),
            )

        # Same PartitionsDefinitions except for start and end dates
        if (
            from_partitions_def.equal_except_for_start_or_end(to_partitions_def)
            and (
                from_partitions_def.start.timestamp() >= to_partitions_def.start.timestamp()
                or from_partitions_subset.first_start.timestamp()
                >= to_partitions_def.start.timestamp()
            )
            and (
                to_partitions_def.end is None
                or (
                    from_partitions_def.end is not None
                    and to_partitions_def.end.timestamp() >= from_partitions_def.end.timestamp()
                )
            )
        ):
            return UpstreamPartitionsResult(
                partitions_subset=from_partitions_subset.with_partitions_def(to_partitions_def),
                required_but_nonexistent_subset=to_partitions_def.empty_subset(),
            )

        # Daily to hourly
        from_last_partition_window = from_partitions_def.get_last_partition_window(current_time)
        to_last_partition_window = to_partitions_def.get_last_partition_window(current_time)
        if (
            from_partitions_def.is_basic_daily
            and to_partitions_def.is_basic_hourly
            and (
                from_partitions_def.start.timestamp() >= to_partitions_def.start.timestamp()
                or from_partitions_subset.first_start.timestamp()
                >= to_partitions_def.start.timestamp()
            )
            and (
                from_last_partition_window is not None
                and to_last_partition_window is not None
                and from_last_partition_window.end.timestamp()
                <= to_last_partition_window.end.timestamp()
            )
        ):
            return UpstreamPartitionsResult(
                partitions_subset=TimeWindowPartitionsSubset(
                    partitions_def=to_partitions_def,
                    num_partitions=None,
                    included_time_windows=from_partitions_subset.included_time_windows,
                ),
                required_but_nonexistent_subset=to_partitions_def.empty_subset(),
            )

        # The subset we're mapping from doesn't exist in the PartitionsDefinition we're mapping to
        if from_partitions_subset.cheap_ends_before(
            to_partitions_def.start, to_partitions_def.cron_schedule
        ):
            if self.allow_nonexistent_upstream_partitions:
                required_but_nonexistent_subset = to_partitions_def.empty_subset()
            else:
                required_but_nonexistent_subset = to_partitions_def.empty_subset()
                for time_window in from_partitions_subset.included_time_windows:
                    required_but_nonexistent_subset = (
                        required_but_nonexistent_subset
                        | to_partitions_def.get_partition_subset_in_time_window(
                            time_window.to_public_time_window()
                        )
                    )

            return UpstreamPartitionsResult(
                partitions_subset=to_partitions_def.empty_subset(),
                required_but_nonexistent_subset=required_but_nonexistent_subset,
            )

        return None


def _offsetted_datetime_with_bounds(
    partitions_def: TimeWindowPartitionsDefinition,
    dt: datetime,
    offset: int,
    bounds_window: Optional[TimeWindow],
) -> datetime:
    offsetted_dt = _offsetted_datetime(partitions_def, dt, offset)

    # Don't allow offsetting to push the windows out of the bounds of the target
    # PartitionsDefinition
    if bounds_window is not None:
        if offset < 0:
            offsetted_dt = max(bounds_window.start, offsetted_dt, key=lambda d: d.timestamp())

        if offset > 0:
            offsetted_dt = min(bounds_window.end, offsetted_dt, key=lambda d: d.timestamp())

    return offsetted_dt


def _offsetted_datetime(
    partitions_def: TimeWindowPartitionsDefinition, dt: datetime, offset: int
) -> datetime:
    if partitions_def.is_basic_daily and offset != 0:
        result = dt + timedelta(days=offset)
        # Handle DST transitions
        if result.hour == 23:
            result = result + timedelta(hours=1)
        elif result.hour == 1:
            result = result - timedelta(hours=1)
        elif result.hour != 0:
            raise Exception(f"Unexpected time after adding day offset {result}")
        return result

    elif partitions_def.is_basic_hourly and offset != 0:
        return add_absolute_time(dt, hours=offset)

    result = dt
    for _ in range(abs(offset)):
        if offset < 0:
            prev_window = cast(
                TimeWindow,
                partitions_def.get_prev_partition_window(result, respect_bounds=False),
            )
            result = prev_window.start
        else:
            next_window = cast(
                TimeWindow,
                partitions_def.get_next_partition_window(result, respect_bounds=False),
            )
            result = next_window.end

    return result



import inspect
from collections.abc import Iterator, Mapping, Sequence, Set
from typing import TYPE_CHECKING, AbstractSet, Any, Callable, Optional, Union, cast  # noqa: UP035

from typing_extensions import TypeAlias, get_args, get_origin

import dagster._check as check
from dagster._annotations import deprecated, deprecated_param, public
from dagster._config.config_schema import UserConfigSchema
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.definition_config_schema import (
    IDefinitionConfigSchema,
    convert_user_facing_definition_config_schema,
)
from dagster._core.definitions.dependency import NodeHandle, NodeInputHandle, NodeOutputHandle
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.inference import infer_output_props
from dagster._core.definitions.input import In, InputDefinition
from dagster._core.definitions.node_definition import NodeDefinition
from dagster._core.definitions.op_invocation import direct_invocation_result
from dagster._core.definitions.output import Out, OutputDefinition
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_requirement import (
    InputManagerRequirement,
    OpDefinitionResourceRequirement,
    OutputManagerRequirement,
    ResourceRequirement,
)
from dagster._core.definitions.result import MaterializeResult, ObserveResult
from dagster._core.definitions.utils import DEFAULT_IO_MANAGER_KEY, check_valid_chars
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvariantViolationError,
)
from dagster._core.storage.tags import GLOBAL_CONCURRENCY_TAG
from dagster._core.types.dagster_type import DagsterType, DagsterTypeKind
from dagster._utils import IHasInternalInit
from dagster._utils.warnings import normalize_renamed_param

if TYPE_CHECKING:
    from dagster._core.definitions.asset_layer import AssetLayer
    from dagster._core.definitions.composition import PendingNodeInvocation
    from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction

OpComputeFunction: TypeAlias = Callable[..., Any]


@deprecated_param(
    param="version",
    breaking_version="2.0",
    additional_warn_text="Use `code_version` instead.",
)
class OpDefinition(NodeDefinition, IHasInternalInit):
    """Defines an op, the functional unit of user-defined computation.

    End users should prefer the :func:`@op <op>` decorator. OpDefinition is generally intended to be
    used by framework authors or for programatically generated ops.

    Args:
        name (str): Name of the op. Must be unique within any :py:class:`GraphDefinition` or
            :py:class:`JobDefinition` that contains the op.
        input_defs (List[InputDefinition]): Inputs of the op.
        compute_fn (Callable): The core of the op, the function that performs the actual
            computation. The signature of this function is determined by ``input_defs``, and
            optionally, an injected first argument, ``context``, a collection of information
            provided by the system.

            This function will be coerced into a generator or an async generator, which must yield
            one :py:class:`Output` for each of the op's ``output_defs``, and additionally may
            yield other types of Dagster events, including :py:class:`AssetMaterialization` and
            :py:class:`ExpectationResult`.
        output_defs (List[OutputDefinition]): Outputs of the op.
        config_schema (Optional[ConfigSchema): The schema for the config. If set, Dagster will check
            that the config provided for the op matches this schema and will fail if it does not. If
            not set, Dagster will accept any config provided for the op.
        description (Optional[str]): Human-readable description of the op.
        tags (Optional[Dict[str, Any]]): Arbitrary metadata for the op. Frameworks may
            expect and require certain metadata to be attached to a op. Users should generally
            not set metadata directly. Values that are not strings will be json encoded and must meet
            the criteria that `json.loads(json.dumps(value)) == value`.
        required_resource_keys (Optional[Set[str]]): Set of resources handles required by this op.
        code_version (Optional[str]): Version of the code encapsulated by the op. If set,
            this is used as a default code version for all outputs.
        retry_policy (Optional[RetryPolicy]): The retry policy for this op.
        pool (Optional[str]): A string that identifies the pool that governs this op's execution.


    Examples:
        .. code-block:: python

            def _add_one(_context, inputs):
                yield Output(inputs["num"] + 1)

            OpDefinition(
                name="add_one",
                ins={"num": In(int)},
                outs={"result": Out(int)},
                compute_fn=_add_one,
            )
    """

    _compute_fn: Union[Callable[..., Any], "DecoratedOpFunction"]
    _config_schema: IDefinitionConfigSchema
    _required_resource_keys: AbstractSet[str]
    _version: Optional[str]
    _retry_policy: Optional[RetryPolicy]
    _pool: Optional[str]

    def __init__(
        self,
        compute_fn: Union[Callable[..., Any], "DecoratedOpFunction"],
        name: str,
        ins: Optional[Mapping[str, In]] = None,
        outs: Optional[Mapping[str, Out]] = None,
        description: Optional[str] = None,
        config_schema: Optional[Union[UserConfigSchema, IDefinitionConfigSchema]] = None,
        required_resource_keys: Optional[AbstractSet[str]] = None,
        tags: Optional[Mapping[str, Any]] = None,
        version: Optional[str] = None,
        retry_policy: Optional[RetryPolicy] = None,
        code_version: Optional[str] = None,
        pool: Optional[str] = None,
    ):
        from dagster._core.definitions.decorators.op_decorator import (
            DecoratedOpFunction,
            resolve_checked_op_fn_inputs,
        )

        ins = check.opt_mapping_param(ins, "ins")
        input_defs = [
            inp.to_definition(name) for name, inp in sorted(ins.items(), key=lambda inp: inp[0])
        ]  # sort so that input definition order is deterministic

        if isinstance(compute_fn, DecoratedOpFunction):
            resolved_input_defs: Sequence[InputDefinition] = resolve_checked_op_fn_inputs(
                decorator_name="@op",
                fn_name=name,
                compute_fn=cast(DecoratedOpFunction, compute_fn),
                explicit_input_defs=input_defs,
                exclude_nothing=True,
            )
            self._compute_fn = compute_fn
            _validate_context_type_hint(self._compute_fn.decorated_fn)
        else:
            resolved_input_defs = input_defs
            self._compute_fn = check.callable_param(compute_fn, "compute_fn")
            _validate_context_type_hint(self._compute_fn)

        code_version = normalize_renamed_param(
            code_version,
            "code_version",
            version,
            "version",
        )
        self._version = code_version

        check.opt_mapping_param(outs, "outs")
        output_defs = _resolve_output_defs_from_outs(
            compute_fn=compute_fn, outs=outs, default_code_version=code_version
        )

        self._config_schema = convert_user_facing_definition_config_schema(config_schema)
        self._required_resource_keys = frozenset(
            check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
        )
        self._retry_policy = check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy)
        self._pool = pool
        pool = _validate_pool(pool, tags)

        positional_inputs = (
            self._compute_fn.positional_inputs()
            if isinstance(self._compute_fn, DecoratedOpFunction)
            else None
        )

        super().__init__(
            name=name,
            input_defs=check.sequence_param(resolved_input_defs, "input_defs", InputDefinition),
            output_defs=check.sequence_param(output_defs, "output_defs", OutputDefinition),
            description=description,
            tags=(check.opt_mapping_param(tags, "tags", key_type=str)),
            positional_inputs=positional_inputs,
        )

    def dagster_internal_init(
        *,
        compute_fn: Union[Callable[..., Any], "DecoratedOpFunction"],
        name: str,
        ins: Optional[Mapping[str, In]],
        outs: Optional[Mapping[str, Out]],
        description: Optional[str],
        config_schema: Optional[Union[UserConfigSchema, IDefinitionConfigSchema]],
        required_resource_keys: Optional[AbstractSet[str]],
        tags: Optional[Mapping[str, Any]],
        version: Optional[str],
        retry_policy: Optional[RetryPolicy],
        code_version: Optional[str],
        pool: Optional[str],
    ) -> "OpDefinition":
        return OpDefinition(
            compute_fn=compute_fn,
            name=name,
            ins=ins,
            outs=outs,
            description=description,
            config_schema=config_schema,
            required_resource_keys=required_resource_keys,
            tags=tags,
            version=version,
            retry_policy=retry_policy,
            code_version=code_version,
            pool=pool,
        )

    @property
    def node_type_str(self) -> str:
        return "op"

    @property
    def is_graph_job_op_node(self) -> bool:
        return True

    @public
    @property
    def name(self) -> str:
        """str: The name of this op."""
        return super().name

    @public
    @property
    def ins(self) -> Mapping[str, In]:
        """Mapping[str, In]: A mapping from input name to the In object that represents that input."""
        return {input_def.name: In.from_definition(input_def) for input_def in self.input_defs}

    @public
    @property
    def outs(self) -> Mapping[str, Out]:
        """Mapping[str, Out]: A mapping from output name to the Out object that represents that output."""
        return {output_def.name: Out.from_definition(output_def) for output_def in self.output_defs}

    @property
    def compute_fn(self) -> Union[Callable[..., Any], "DecoratedOpFunction"]:
        return self._compute_fn

    @public
    @property
    def config_schema(self) -> IDefinitionConfigSchema:
        """IDefinitionConfigSchema: The config schema for this op."""
        return self._config_schema

    @public
    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        """AbstractSet[str]: A set of keys for resources that must be provided to this OpDefinition."""
        return frozenset(self._required_resource_keys)

    @public
    @deprecated(breaking_version="2.0", additional_warn_text="Use `code_version` instead.")
    @property
    def version(self) -> Optional[str]:
        """str: Version of the code encapsulated by the op. If set, this is used as a
        default code version for all outputs.
        """
        return self._version

    @public
    @property
    def retry_policy(self) -> Optional[RetryPolicy]:
        """Optional[RetryPolicy]: The RetryPolicy for this op."""
        return self._retry_policy

    @public
    @property
    def tags(self) -> Mapping[str, str]:
        """Mapping[str, str]: The tags for this op."""
        return super().tags

    @public
    def alias(self, name: str) -> "PendingNodeInvocation":
        """Creates a copy of this op with the given name."""
        return super().alias(name)

    @public
    def tag(self, tags: Optional[Mapping[str, str]]) -> "PendingNodeInvocation":
        """Creates a copy of this op with the given tags."""
        return super().tag(tags)

    @public
    def with_hooks(self, hook_defs: AbstractSet[HookDefinition]) -> "PendingNodeInvocation":
        """Creates a copy of this op with the given hook definitions."""
        return super().with_hooks(hook_defs)

    @public
    def with_retry_policy(self, retry_policy: RetryPolicy) -> "PendingNodeInvocation":
        """Creates a copy of this op with the given retry policy."""
        return super().with_retry_policy(retry_policy)

    @property
    def pool(self) -> Optional[str]:
        """Optional[str]: The concurrency pool for this op."""
        return self._pool

    @property
    def pools(self) -> Set[str]:
        """Optional[str]: The concurrency pools for this op node."""
        return {self._pool} if self._pool else set()

    def is_from_decorator(self) -> bool:
        from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction

        return isinstance(self._compute_fn, DecoratedOpFunction)

    def get_output_annotation(self) -> Any:
        if not self.is_from_decorator():
            raise DagsterInvalidInvocationError(
                f"Attempted to get output annotation for {self.node_type_str} '{self.name}', "
                "which was not constructed from a decorated function."
            )
        return cast("DecoratedOpFunction", self.compute_fn).get_output_annotation()

    def all_dagster_types(self) -> Iterator[DagsterType]:
        yield from self.all_input_output_types()

    def iterate_node_defs(self) -> Iterator[NodeDefinition]:
        yield self

    def iterate_op_defs(self) -> Iterator["OpDefinition"]:
        yield self

    def resolve_output_to_origin(
        self, output_name: str, handle: Optional[NodeHandle]
    ) -> tuple[OutputDefinition, Optional[NodeHandle]]:
        return self.output_def_named(output_name), handle

    def resolve_output_to_origin_op_def(self, output_name: str) -> "OpDefinition":
        return self

    def get_inputs_must_be_resolved_top_level(
        self, asset_layer: "AssetLayer", handle: Optional[NodeHandle] = None
    ) -> Sequence[InputDefinition]:
        handle = cast(NodeHandle, check.inst_param(handle, "handle", NodeHandle))
        unresolveable_input_defs = []
        for input_def in self.input_defs:
            if (
                not input_def.dagster_type.loader
                and not input_def.dagster_type.kind == DagsterTypeKind.NOTHING
                and not input_def.has_default_value
                and not input_def.input_manager_key
            ):
                input_asset_key = asset_layer.asset_key_for_input(handle, input_def.name)
                # If input_asset_key is present, this input can be resolved
                # by a source asset, so input does not need to be resolved
                # at the top level.
                if input_asset_key:
                    continue
                unresolveable_input_defs.append(input_def)
        return unresolveable_input_defs

    def input_has_default(self, input_name: str) -> bool:
        return self.input_def_named(input_name).has_default_value

    def default_value_for_input(self, input_name: str) -> InputDefinition:
        return self.input_def_named(input_name).default_value

    def input_supports_dynamic_output_dep(self, input_name: str) -> bool:
        return True

    def with_replaced_properties(
        self,
        name: str,
        ins: Optional[Mapping[str, In]] = None,
        outs: Optional[Mapping[str, Out]] = None,
        config_schema: Optional[IDefinitionConfigSchema] = None,
        description: Optional[str] = None,
    ) -> "OpDefinition":
        return OpDefinition.dagster_internal_init(
            name=name,
            ins={input_def.name: In.from_definition(input_def) for input_def in self.input_defs}
            if ins is None
            else ins,
            outs={
                output_def.name: Out.from_definition(output_def) for output_def in self.output_defs
            }
            if outs is None
            else outs,
            compute_fn=self.compute_fn,
            config_schema=config_schema or self.config_schema,
            description=description or self.description,
            tags=self.tags,
            required_resource_keys=self.required_resource_keys,
            code_version=self._version,
            retry_policy=self.retry_policy,
            version=None,  # code_version replaces version
            pool=self.pool,
        )

    def copy_for_configured(
        self,
        name: str,
        description: Optional[str],
        config_schema: IDefinitionConfigSchema,
    ) -> "OpDefinition":
        return self.with_replaced_properties(
            name=name,
            description=description,
            config_schema=config_schema,
        )

    def get_resource_requirements(
        self,
        handle: Optional[NodeHandle],
        asset_layer: Optional["AssetLayer"],
    ) -> Iterator[ResourceRequirement]:
        node_description = f"{self.node_type_str} '{handle or self.name}'"
        for resource_key in sorted(list(self.required_resource_keys)):
            yield OpDefinitionResourceRequirement(
                key=resource_key, node_description=node_description
            )
        for input_def in self.input_defs:
            if input_def.input_manager_key:
                yield InputManagerRequirement(
                    key=input_def.input_manager_key,
                    node_description=node_description,
                    input_name=input_def.name,
                    root_input=False,
                )
            elif asset_layer and handle:
                input_asset_key = asset_layer.asset_key_for_input(handle, input_def.name)
                if input_asset_key:
                    io_manager_key = (
                        asset_layer.get(input_asset_key).io_manager_key
                        if asset_layer.has(input_asset_key)
                        else DEFAULT_IO_MANAGER_KEY
                    )
                    yield InputManagerRequirement(
                        key=io_manager_key,
                        node_description=node_description,
                        input_name=input_def.name,
                        root_input=False,
                    )

        for output_def in self.output_defs:
            yield OutputManagerRequirement(
                key=output_def.io_manager_key,
                node_description=node_description,
                output_name=output_def.name,
            )

    def resolve_input_to_destinations(
        self, input_handle: NodeInputHandle
    ) -> Sequence[NodeInputHandle]:
        return [input_handle]

    def resolve_output_to_destinations(
        self, output_name: str, handle: Optional[NodeHandle]
    ) -> Sequence[NodeInputHandle]:
        return []

    def __call__(self, *args, **kwargs) -> Any:
        from dagster._core.definitions.composition import is_in_composition

        if is_in_composition():
            return super().__call__(*args, **kwargs)

        return direct_invocation_result(self, *args, **kwargs)

    def get_op_handles(self, parent: NodeHandle) -> AbstractSet[NodeHandle]:
        return {parent}

    def get_op_output_handles(self, parent: Optional[NodeHandle]) -> AbstractSet[NodeOutputHandle]:
        return {
            NodeOutputHandle(node_handle=parent, output_name=output_def.name)
            for output_def in self.output_defs
        }


def _resolve_output_defs_from_outs(
    compute_fn: Union[Callable[..., Any], "DecoratedOpFunction"],
    outs: Optional[Mapping[str, Out]],
    default_code_version: Optional[str],
) -> Sequence[OutputDefinition]:
    from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction

    if isinstance(compute_fn, DecoratedOpFunction):
        inferred_output_props = infer_output_props(compute_fn.decorated_fn)
        annotation = inferred_output_props.annotation
        description = inferred_output_props.description
    else:
        inferred_output_props = None
        annotation = inspect.Parameter.empty
        description = None

    if outs is None:
        return [OutputDefinition.create_from_inferred(inferred_output_props, default_code_version)]

    # If only a single entry has been provided to the out dict, then slurp the
    # annotation into the entry.
    if len(outs) == 1:
        name = next(iter(outs.keys()))
        only_out = outs[name]
        return [only_out.to_definition(annotation, name, description, default_code_version)]

    # If multiple outputs...

    # Note: we don't provide description when using multiple outputs. Introspection
    # is challenging when faced with multiple outputs.

    # ... and no annotation, use empty for each output annotation
    if annotation == inspect.Parameter.empty:
        return [
            out.to_definition(
                annotation_type=inspect.Parameter.empty,
                name=name,
                description=None,
                code_version=default_code_version,
            )
            for (name, out) in outs.items()
        ]

    # ... or if a single result object type, use None for each output annotation
    if _is_result_object_type(annotation):
        # this can happen for example when there are outputs for checks
        # that get reported via a singular MaterializeResult
        return [
            out.to_definition(
                annotation_type=type(None),
                name=name,
                description=None,
                code_version=default_code_version,
            )
            for (name, out) in outs.items()
        ]

    # ... otherwise we expect to have a tuple with entries...
    if get_origin(annotation) != tuple:
        raise DagsterInvariantViolationError(
            "Expected Tuple annotation for multiple outputs, but received non-tuple annotation."
        )
    subtypes = get_args(annotation)

    # ... if they are all result object entries use None
    if len(subtypes) > 0 and all(_is_result_object_type(t) for t in subtypes):
        # the counts of subtypes and outputs may not align due to checks results
        # being passed via MaterializeResult similar to above.
        return [
            out.to_definition(
                annotation_type=type(None),
                name=name,
                description=None,
                code_version=default_code_version,
            )
            for (name, out) in outs.items()
        ]

    # ... otherwise they should align with outputs
    if len(subtypes) != len(outs):
        raise DagsterInvariantViolationError(
            "Expected Tuple annotation to have number of entries matching the "
            f"number of outputs for more than one output. Expected {len(outs)} "
            f"outputs but annotation has {len(subtypes)}."
        )
    return [
        cur_out.to_definition(
            annotation_type=subtypes[idx],
            name=name,
            description=None,
            code_version=default_code_version,
        )
        for idx, (name, cur_out) in enumerate(outs.items())
    ]


def _validate_context_type_hint(fn):
    from inspect import _empty as EmptyAnnotation

    from dagster._core.decorator_utils import get_function_params
    from dagster._core.definitions.decorators.op_decorator import is_context_provided
    from dagster._core.execution.context.compute import (
        AssetCheckExecutionContext,
        AssetExecutionContext,
        OpExecutionContext,
    )

    params = get_function_params(fn)
    if is_context_provided(params):
        if params[0].annotation not in [
            AssetExecutionContext,
            OpExecutionContext,
            EmptyAnnotation,
            AssetCheckExecutionContext,
        ]:
            raise DagsterInvalidDefinitionError(
                f"Cannot annotate `context` parameter with type {params[0].annotation}. `context`"
                " must be annotated with AssetExecutionContext, AssetCheckExecutionContext, OpExecutionContext, or left blank."
            )


def _is_result_object_type(ttype):
    # Is this type special result object type
    return ttype in (MaterializeResult, ObserveResult, AssetCheckResult)


def _validate_pool(pool, tags):
    check.opt_str_param(pool, "pool")
    if not pool:
        return None

    check_valid_chars(pool)

    tags = check.opt_mapping_param(tags, "tags")
    tag_concurrency_key = tags.get(GLOBAL_CONCURRENCY_TAG)
    if tag_concurrency_key and pool != tag_concurrency_key:
        raise DagsterInvalidDefinitionError(
            f'Pool "{pool}" conflicts with the concurrency key tag "{tag_concurrency_key}".'
        )

    return pool



from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from inspect import _empty as EmptyAnnotation
from typing import Optional, Union

from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.execution.context.asset_check_execution_context import AssetCheckExecutionContext
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.execution.context.system import StepExecutionContext

ExecutionContextTypes = Union[OpExecutionContext, AssetExecutionContext, AssetCheckExecutionContext]


@contextmanager
def enter_execution_context(
    step_context: StepExecutionContext,
) -> Iterator[ExecutionContextTypes]:
    """Get the correct context based on the type of step (op or asset) and the user provided context
    type annotation. Follows these rules.

    step type     annotation                   result
    asset         AssetExecutionContext        AssetExecutionContext
    asset         AssetCheckExecutionContext   Error - not an asset check
    asset         OpExecutionContext           OpExecutionContext
    asset         None                         AssetExecutionContext
    op            AssetExecutionContext        Error - we cannot init an AssetExecutionContextext w/o an AssetsDefinition
    op            AssetCheckExecutionContext   Error - not an asset check
    op            OpExecutionContext           OpExecutionContext
    op            None                         OpExecutionContext
    asset_check   AssetCheckExecutionContext   AssetCheckExecutionContext
    asset_check   AssetExecutionContext        Error - not an asset
    asset_check   OpExecutionContext           OpExecutionContext
    asset_check   None                         AssetCheckExecutionContext

    For ops in graph-backed assets
    step type     annotation                   result
    op            AssetExecutionContext        AssetExecutionContext
    op            AssetCheckExecutionContext   Error - not an asset check
    op            OpExecutionContext           OpExecutionContext
    op            None                         OpExecutionContext

    """
    is_sda_step = step_context.is_sda_step
    is_op_in_graph_asset = step_context.is_in_graph_asset
    asset_check_only_step = (
        step_context.is_asset_check_step and not is_sda_step and not is_op_in_graph_asset
    )
    context_annotation = EmptyAnnotation
    compute_fn = step_context.op_def._compute_fn  # noqa: SLF001
    compute_fn = (
        compute_fn
        if isinstance(compute_fn, DecoratedOpFunction)
        else DecoratedOpFunction(compute_fn)
    )
    if compute_fn.has_context_arg():
        context_param = compute_fn.get_context_arg()
        context_annotation = context_param.annotation

    if context_annotation is AssetCheckExecutionContext and not asset_check_only_step:
        if is_sda_step:
            raise DagsterInvalidDefinitionError(
                "Cannot annotate @asset `context` parameter with type AssetCheckExecutionContext. "
                "`context` must be annotated with AssetExecutionContext, OpExecutionContext, or left blank."
            )
        else:
            raise DagsterInvalidDefinitionError(
                "Cannot annotate @op `context` parameter with type AssetCheckExecutionContext. "
                "`context` must be annotated with OpExecutionContext, AssetExecutionContext (if part of "
                "a graph-backed-asset) or left blank."
            )

    if context_annotation is AssetExecutionContext:
        if asset_check_only_step:
            raise DagsterInvalidDefinitionError(
                "Cannot annotate @asset_check `context` parameter with type AssetExecutionContext. "
                "`context` must be annotated with AssetCheckExecutionContext, OpExecutionContext, or left blank."
            )

        # It would be nice to do this check at definition time, rather than at run time, but we don't
        # know if the op is part of an op job or a graph-backed asset until we have the step execution context
        if not is_sda_step and not is_op_in_graph_asset:
            # AssetExecutionContext requires an AssetsDefinition during init, so an op in an op job
            # cannot be annotated with AssetExecutionContext
            raise DagsterInvalidDefinitionError(
                "Cannot annotate @op `context` parameter with type AssetExecutionContext unless the"
                " op is part of a graph-backed asset. `context` must be annotated with"
                " OpExecutionContext, or left blank."
            )

    # default to AssetCheckExecutionContext in @asset_checks
    if asset_check_only_step and context_annotation is not OpExecutionContext:
        asset_ctx = AssetCheckExecutionContext(
            op_execution_context=OpExecutionContext(step_context)
        )
    else:
        asset_ctx = AssetExecutionContext(op_execution_context=OpExecutionContext(step_context))

    asset_token = current_execution_context.set(asset_ctx)

    try:
        if context_annotation is EmptyAnnotation:
            # if no type hint has been given, default to:
            # * AssetExecutionContext for sda steps not in graph-backed assets, and asset_checks
            # * OpExecutionContext for non sda steps
            # * OpExecutionContext for ops in graph-backed assets
            if asset_check_only_step:
                yield asset_ctx
            elif is_op_in_graph_asset or not is_sda_step:
                yield asset_ctx.op_execution_context
            else:
                yield asset_ctx
        elif (
            context_annotation is AssetExecutionContext
            or context_annotation is AssetCheckExecutionContext
        ):
            yield asset_ctx
        else:
            yield asset_ctx.op_execution_context
    finally:
        current_execution_context.reset(asset_token)


current_execution_context: ContextVar[
    Optional[Union[AssetExecutionContext, AssetCheckExecutionContext]]
] = ContextVar("current_execution_context", default=None)



from collections.abc import Iterable, Mapping, Sequence
from typing import AbstractSet, Any, Callable, NamedTuple, Optional, Union, overload  # noqa: UP035

import dagster._check as check
from dagster._annotations import beta_param, hidden_param, only_allow_hidden_params_in_kwargs
from dagster._config.config_schema import UserConfigSchema
from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.asset_dep import (
    AssetDep,
    CoercibleToAssetDep,
    coerce_to_deps_and_check_duplicates,
)
from dagster._core.definitions.asset_in import AssetIn
from dagster._core.definitions.asset_out import AssetOut
from dagster._core.definitions.asset_spec import AssetExecutionType, AssetSpec, validate_kind_tags
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy, BackfillPolicyType
from dagster._core.definitions.config import ConfigMapping
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.decorators.decorator_assets_definition_builder import (
    DecoratorAssetsDefinitionBuilder,
    DecoratorAssetsDefinitionBuilderArgs,
    build_and_validate_named_ins,
    build_named_outs,
    create_check_specs_by_output_name,
    validate_and_assign_output_names_to_check_specs,
)
from dagster._core.definitions.decorators.graph_decorator import graph
from dagster._core.definitions.events import (
    AssetKey,
    CoercibleToAssetKey,
    CoercibleToAssetKeyPrefix,
)
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.input import GraphIn
from dagster._core.definitions.metadata import ArbitraryMetadataMapping, RawMetadataMapping
from dagster._core.definitions.output import GraphOut
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.utils import (
    DEFAULT_IO_MANAGER_KEY,
    DEFAULT_OUTPUT,
    NoValueSentinel,
    resolve_automation_condition,
)
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.storage.tags import KIND_PREFIX
from dagster._core.types.dagster_type import DagsterType
from dagster._utils.tags import normalize_tags
from dagster._utils.warnings import disable_dagster_warnings


@overload
def asset(
    compute_fn: Callable[..., Any],
    **kwargs,
) -> AssetsDefinition: ...


@overload
def asset(
    *,
    name: Optional[str] = ...,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = ...,
    deps: Optional[Iterable[CoercibleToAssetDep]] = ...,
    metadata: Optional[Mapping[str, Any]] = ...,
    tags: Optional[Mapping[str, str]] = ...,
    description: Optional[str] = ...,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[AbstractSet[str]] = ...,
    resource_defs: Optional[Mapping[str, object]] = ...,
    io_manager_def: Optional[object] = ...,
    io_manager_key: Optional[str] = ...,
    dagster_type: Optional[DagsterType] = ...,
    partitions_def: Optional[PartitionsDefinition] = ...,
    op_tags: Optional[Mapping[str, Any]] = ...,
    group_name: Optional[str] = ...,
    output_required: bool = ...,
    automation_condition: Optional[AutomationCondition] = ...,
    backfill_policy: Optional[BackfillPolicy] = ...,
    retry_policy: Optional[RetryPolicy] = ...,
    code_version: Optional[str] = ...,
    key: Optional[CoercibleToAssetKey] = None,
    check_specs: Optional[Sequence[AssetCheckSpec]] = ...,
    owners: Optional[Sequence[str]] = ...,
    kinds: Optional[AbstractSet[str]] = ...,
    pool: Optional[str] = ...,
    **kwargs,
) -> Callable[[Callable[..., Any]], AssetsDefinition]: ...


def _validate_hidden_non_argument_dep_param(
    non_argument_deps: Any,
) -> Optional[Union[set[AssetKey], set[str]]]:
    if non_argument_deps is None:
        return non_argument_deps

    if not isinstance(non_argument_deps, set):
        check.failed("non_arguments_deps must be a set if not None")

    assert isinstance(non_argument_deps, set)

    check.set_param(non_argument_deps, "non_argument_deps", of_type=(str, AssetKey))

    check.invariant(
        all(isinstance(dep, str) for dep in non_argument_deps)
        or all(isinstance(dep, AssetKey) for dep in non_argument_deps),
    )

    return non_argument_deps


@beta_param(param="resource_defs")
@beta_param(param="io_manager_def")
@beta_param(param="backfill_policy")
@hidden_param(
    param="non_argument_deps",
    breaking_version="2.0.0",
    additional_warn_text="use `deps` instead.",
)
@hidden_param(
    param="auto_materialize_policy",
    breaking_version="1.10.0",
    additional_warn_text="use `automation_condition` instead.",
)
@hidden_param(
    param="freshness_policy",
    breaking_version="1.10.0",
    additional_warn_text="use freshness checks instead.",
)
@hidden_param(
    param="compute_kind",
    emit_runtime_warning=False,
    breaking_version="1.10.0",
)
def asset(
    compute_fn: Optional[Callable[..., Any]] = None,
    *,
    name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    deps: Optional[Iterable[CoercibleToAssetDep]] = None,
    metadata: Optional[ArbitraryMetadataMapping] = None,
    tags: Optional[Mapping[str, str]] = None,
    description: Optional[str] = None,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
    resource_defs: Optional[Mapping[str, object]] = None,
    io_manager_def: Optional[object] = None,
    io_manager_key: Optional[str] = None,
    dagster_type: Optional[DagsterType] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    group_name: Optional[str] = None,
    output_required: bool = True,
    automation_condition: Optional[AutomationCondition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    retry_policy: Optional[RetryPolicy] = None,
    code_version: Optional[str] = None,
    key: Optional[CoercibleToAssetKey] = None,
    check_specs: Optional[Sequence[AssetCheckSpec]] = None,
    owners: Optional[Sequence[str]] = None,
    kinds: Optional[AbstractSet[str]] = None,
    pool: Optional[str] = None,
    **kwargs,
) -> Union[AssetsDefinition, Callable[[Callable[..., Any]], AssetsDefinition]]:
    """Create a definition for how to compute an asset.

    A software-defined asset is the combination of:
      1. An asset key, e.g. the name of a table.
      2. A function, which can be run to compute the contents of the asset.
      3. A set of upstream assets that are provided as inputs to the function when computing the asset.

    Unlike an op, whose dependencies are determined by the graph it lives inside, an asset knows
    about the upstream assets it depends on. The upstream assets are inferred from the arguments
    to the decorated function. The name of the argument designates the name of the upstream asset.

    An asset has an op inside it to represent the function that computes it. The name of the op
    will be the segments of the asset key, separated by double-underscores.

    Args:
        name (Optional[str]): The name of the asset.  If not provided, defaults to the name of the
            decorated function. The asset's name must be a valid name in dagster (ie only contains
            letters, numbers, and _) and may not contain python reserved keywords.
        key_prefix (Optional[Union[str, Sequence[str]]]): If provided, the asset's key is the
            concatenation of the key_prefix and the asset's name, which defaults to the name of
            the decorated function. Each item in key_prefix must be a valid name in dagster (ie only
            contains letters, numbers, and _) and may not contain python reserved keywords.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        deps (Optional[Sequence[Union[AssetDep, AssetsDefinition, SourceAsset, AssetKey, str]]]):
            The assets that are upstream dependencies, but do not correspond to a parameter of the
            decorated function. If the AssetsDefinition for a multi_asset is provided, dependencies on
            all assets created by the multi_asset will be created.
        config_schema (Optional[ConfigSchema): The configuration schema for the asset's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        metadata (Optional[Dict[str, Any]]): A dict of metadata entries for the asset.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
        required_resource_keys (Optional[Set[str]]): Set of resource handles required by the op.
        io_manager_key (Optional[str]): The resource key of the IOManager used
            for storing the output of the op as an asset, and for loading it in downstream ops
            (default: "io_manager"). Only one of io_manager_key and io_manager_def can be provided.
        io_manager_def (Optional[object]): (Beta) The IOManager used for
            storing the output of the op as an asset,  and for loading it in
            downstream ops. Only one of io_manager_def and io_manager_key can be provided.
        dagster_type (Optional[DagsterType]): Allows specifying type validation functions that
            will be executed on the output of the decorated function after it runs.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that computes the asset.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If not provided,
            the name "default" is used.
        resource_defs (Optional[Mapping[str, object]]):
            (Beta) A mapping of resource keys to resources. These resources
            will be initialized during execution, and can be accessed from the
            context within the body of the function.
        output_required (bool): Whether the decorated function will always materialize an asset.
            Defaults to True. If False, the function can conditionally not yield a result. If
            no result is yielded, no output will be materialized to storage and downstream
            assets will not be materialized.
        automation_condition (AutomationCondition): A condition describing when Dagster should materialize this asset.
        backfill_policy (BackfillPolicy): (Beta) Configure Dagster to backfill this asset according to its
            BackfillPolicy.
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that computes the asset.
        code_version (Optional[str]): Version of the code that generates this asset. In
            general, versions should be set only for code that deterministically produces the same
            output when given the same inputs.
        check_specs (Optional[Sequence[AssetCheckSpec]]): Specs for asset checks that
            execute in the decorated function after materializing the asset.
        key (Optional[CoeercibleToAssetKey]): The key for this asset. If provided, cannot specify key_prefix or name.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the asset. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        kinds (Optional[Set[str]]): A list of strings representing the kinds of the asset. These
            will be made visible in the Dagster UI.
        pool (Optional[str]): A string that identifies the concurrency pool that governs this asset's execution.
        non_argument_deps (Optional[Union[Set[AssetKey], Set[str]]]): Deprecated, use deps instead.
            Set of asset keys that are upstream dependencies, but do not pass an input to the asset.
            Hidden parameter not exposed in the decorator signature, but passed in kwargs.

    Examples:
        .. code-block:: python

            @asset
            def my_upstream_asset() -> int:
                return 5

            @asset
            def my_asset(my_upstream_asset: int) -> int:
                return my_upstream_asset + 1

            should_materialize = True

            @asset(output_required=False)
            def conditional_asset():
                if should_materialize:
                    yield Output(5)  # you must `yield`, not `return`, the result

            # Will also only materialize if `should_materialize` is `True`
            @asset
            def downstream_asset(conditional_asset):
                return conditional_asset + 1

    """
    compute_kind = check.opt_str_param(kwargs.get("compute_kind"), "compute_kind")
    required_resource_keys = check.opt_set_param(required_resource_keys, "required_resource_keys")
    upstream_asset_deps = _deps_and_non_argument_deps_to_asset_deps(
        deps=deps,
        non_argument_deps=_validate_hidden_non_argument_dep_param(kwargs.get("non_argument_deps")),
    )
    resource_defs = dict(check.opt_mapping_param(resource_defs, "resource_defs"))

    if compute_kind and kinds:
        raise DagsterInvalidDefinitionError(
            "Cannot specify compute_kind and kinds on the @asset decorator."
        )
    validate_kind_tags(kinds)
    tags_with_kinds = {
        **(normalize_tags(tags, strict=True)),
        **{f"{KIND_PREFIX}{kind}": "" for kind in kinds or []},
    }

    only_allow_hidden_params_in_kwargs(asset, kwargs)

    args = AssetDecoratorArgs(
        name=name,
        key_prefix=key_prefix,
        ins=ins or {},
        deps=upstream_asset_deps or [],
        metadata=metadata,
        tags=tags_with_kinds,
        description=description,
        config_schema=config_schema,
        required_resource_keys=required_resource_keys,
        resource_defs=resource_defs,
        io_manager_key=io_manager_key,
        io_manager_def=io_manager_def,
        compute_kind=compute_kind,
        dagster_type=dagster_type,
        partitions_def=partitions_def,
        op_tags=op_tags,
        group_name=group_name,
        output_required=output_required,
        freshness_policy=kwargs.get("freshness_policy"),
        automation_condition=resolve_automation_condition(
            automation_condition, kwargs.get("auto_materialize_policy")
        ),
        backfill_policy=backfill_policy,
        retry_policy=retry_policy,
        code_version=code_version,
        check_specs=check_specs,
        key=key,
        owners=owners,
        pool=pool,
    )

    if compute_fn is not None:
        return create_assets_def_from_fn_and_decorator_args(args, compute_fn)

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        check.invariant(
            not (io_manager_key and io_manager_def),
            "Both io_manager_key and io_manager_def were provided to `@asset` decorator. Please"
            " provide one or the other. ",
        )
        return create_assets_def_from_fn_and_decorator_args(args, fn)

    return inner


def resolve_asset_key_and_name_for_decorator(
    *,
    key: Optional[CoercibleToAssetKey],
    key_prefix: Optional[CoercibleToAssetKeyPrefix],
    name: Optional[str],
    decorator_name: str,
    fn: Callable[..., Any],
) -> tuple[AssetKey, str]:
    if (name or key_prefix) and key:
        raise DagsterInvalidDefinitionError(
            f"Cannot specify a name or key prefix for {decorator_name} when the key"
            " argument is provided."
        )
    key_prefix_list = [key_prefix] if isinstance(key_prefix, str) else key_prefix
    key = AssetKey.from_coercible(key) if key else None
    assigned_name = name or fn.__name__
    return (
        (
            # the filter here appears unnecessary per typing, but this exists
            # historically so keeping it here to be conservative in case users
            # can get Nones into the key_prefix_list somehow
            AssetKey(list(filter(None, [*(key_prefix_list or []), assigned_name])))
            if not key
            else key
        ),
        assigned_name,
    )


class AssetDecoratorArgs(NamedTuple):
    required_resource_keys: AbstractSet[str]
    name: Optional[str]
    key_prefix: Optional[CoercibleToAssetKeyPrefix]
    ins: Mapping[str, AssetIn]
    deps: Iterable[AssetDep]
    metadata: Optional[ArbitraryMetadataMapping]
    tags: Optional[Mapping[str, str]]
    description: Optional[str]
    config_schema: Optional[UserConfigSchema]
    resource_defs: dict[str, object]
    io_manager_key: Optional[str]
    io_manager_def: Optional[object]
    compute_kind: Optional[str]
    dagster_type: Optional[DagsterType]
    partitions_def: Optional[PartitionsDefinition]
    op_tags: Optional[Mapping[str, Any]]
    group_name: Optional[str]
    output_required: bool
    freshness_policy: Optional[FreshnessPolicy]
    automation_condition: Optional[AutomationCondition]
    backfill_policy: Optional[BackfillPolicy]
    retry_policy: Optional[RetryPolicy]
    code_version: Optional[str]
    key: Optional[CoercibleToAssetKey]
    check_specs: Optional[Sequence[AssetCheckSpec]]
    owners: Optional[Sequence[str]]
    pool: Optional[str]


class ResourceRelatedState(NamedTuple):
    io_manager_def: Optional[object]
    io_manager_key: Optional[str]
    resources: Mapping[str, object]
    out_asset_key: AssetKey

    @property
    def op_resource_defs(self) -> Mapping[str, ResourceDefinition]:
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        return wrap_resources_for_execution(self.resources)

    @property
    def resolved_io_manager_key(self) -> str:
        if self.io_manager_def:
            return (
                self.io_manager_key
                if self.io_manager_key
                else self.out_asset_key.to_python_identifier("io_manager")
            )
        else:
            return self.io_manager_key if self.io_manager_key else DEFAULT_IO_MANAGER_KEY

    @property
    def asset_resource_defs(self) -> Mapping[str, ResourceDefinition]:
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        # If these was no io_manager def directly passed in we can just wrap
        # the explicitly provided resource defs
        if not self.io_manager_def:
            return wrap_resources_for_execution(self.resources)

        io_manager_key = self.resolved_io_manager_key
        io_manager_def = self.io_manager_def
        if io_manager_key in self.resources and self.resources[io_manager_key] != io_manager_def:
            raise DagsterInvalidDefinitionError(
                f"Provided conflicting definitions for io manager key '{io_manager_key}'."
                " Please provide only one definition per key."
            )

        return wrap_resources_for_execution({**self.resources, **{io_manager_key: io_manager_def}})


def create_assets_def_from_fn_and_decorator_args(
    args: AssetDecoratorArgs, fn: Callable[..., Any]
) -> AssetsDefinition:
    from dagster._config.pythonic_config import validate_resource_annotated_function

    validate_resource_annotated_function(fn)

    out_asset_key, asset_name = resolve_asset_key_and_name_for_decorator(
        key=args.key,
        key_prefix=args.key_prefix,
        name=args.name,
        fn=fn,
        decorator_name="@asset",
    )

    resource_related_state = ResourceRelatedState(
        io_manager_def=args.io_manager_def,
        io_manager_key=args.io_manager_key,
        resources=args.resource_defs,
        out_asset_key=out_asset_key,
    )

    with disable_dagster_warnings():
        # check backfill policy is BackfillPolicyType.SINGLE_RUN for non-partitioned asset
        if args.partitions_def is None:
            check.param_invariant(
                (
                    args.backfill_policy.policy_type is BackfillPolicyType.SINGLE_RUN
                    if args.backfill_policy
                    else True
                ),
                "backfill_policy",
                "Non partitioned asset can only have single run backfill policy",
            )

    with disable_dagster_warnings():
        builder_args = DecoratorAssetsDefinitionBuilderArgs(
            name=args.name,
            op_description=args.description,
            check_specs_by_output_name=create_check_specs_by_output_name(args.check_specs),
            group_name=args.group_name,
            partitions_def=args.partitions_def,
            retry_policy=args.retry_policy,
            code_version=args.code_version,
            op_tags=args.op_tags,
            config_schema=args.config_schema,
            compute_kind=args.compute_kind,
            required_resource_keys=args.required_resource_keys,
            op_def_resource_defs=resource_related_state.op_resource_defs,
            assets_def_resource_defs=resource_related_state.asset_resource_defs,
            backfill_policy=args.backfill_policy,
            asset_out_map={
                DEFAULT_OUTPUT: AssetOut(
                    key=out_asset_key,
                    metadata=args.metadata,
                    description=args.description,
                    is_required=args.output_required,
                    io_manager_key=resource_related_state.resolved_io_manager_key,
                    dagster_type=args.dagster_type if args.dagster_type else NoValueSentinel,
                    group_name=args.group_name,
                    code_version=args.code_version,
                    freshness_policy=args.freshness_policy,
                    automation_condition=args.automation_condition,
                    backfill_policy=args.backfill_policy,
                    owners=args.owners,
                    tags=normalize_tags(args.tags or {}, strict=True),
                )
            },
            upstream_asset_deps=args.deps,
            asset_in_map=args.ins,
            # We will not be using specs to construct here
            # because they are assumption about output names. Non-spec
            # construction path assumptions apply here
            specs=[],
            # no internal asset deps
            asset_deps={},
            can_subset=False,
            decorator_name="@asset",
            execution_type=AssetExecutionType.MATERIALIZATION,
            pool=args.pool,
        )

        builder = DecoratorAssetsDefinitionBuilder.from_asset_outs_in_asset_centric_decorator(
            fn=fn,
            op_name=out_asset_key.to_python_identifier(),
            asset_in_map=builder_args.asset_in_map,
            asset_out_map=builder_args.asset_out_map,
            asset_deps=builder_args.asset_deps,
            upstream_asset_deps=builder_args.upstream_asset_deps,
            passed_args=builder_args,
        )

    return builder.create_assets_definition()


@beta_param(param="resource_defs")
@hidden_param(
    param="non_argument_deps",
    breaking_version="2.0.0",
    additional_warn_text="use `deps` instead.",
)
@hidden_param(
    param="compute_kind",
    emit_runtime_warning=False,
    breaking_version="1.10.0",
)
def multi_asset(
    *,
    outs: Optional[Mapping[str, AssetOut]] = None,
    name: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    deps: Optional[Iterable[CoercibleToAssetDep]] = None,
    description: Optional[str] = None,
    config_schema: Optional[UserConfigSchema] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
    internal_asset_deps: Optional[Mapping[str, set[AssetKey]]] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    op_tags: Optional[Mapping[str, Any]] = None,
    can_subset: bool = False,
    resource_defs: Optional[Mapping[str, object]] = None,
    group_name: Optional[str] = None,
    retry_policy: Optional[RetryPolicy] = None,
    code_version: Optional[str] = None,
    specs: Optional[Sequence[AssetSpec]] = None,
    check_specs: Optional[Sequence[AssetCheckSpec]] = None,
    pool: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    """Create a combined definition of multiple assets that are computed using the same op and same
    upstream assets.

    Each argument to the decorated function references an upstream asset that this asset depends on.
    The name of the argument designates the name of the upstream asset.

    You can set I/O managers keys, auto-materialize policies, freshness policies, group names, etc.
    on an individual asset within the multi-asset by attaching them to the :py:class:`AssetOut`
    corresponding to that asset in the `outs` parameter.

    Args:
        name (Optional[str]): The name of the op.
        outs: (Optional[Dict[str, AssetOut]]): The AssetOuts representing the assets materialized by
            this function. AssetOuts detail the output, IO management, and core asset properties.
            This argument is required except when AssetSpecs are used.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        deps (Optional[Sequence[Union[AssetsDefinition, SourceAsset, AssetKey, str]]]):
            The assets that are upstream dependencies, but do not correspond to a parameter of the
            decorated function. If the AssetsDefinition for a multi_asset is provided, dependencies on
            all assets created by the multi_asset will be created.
        config_schema (Optional[ConfigSchema): The configuration schema for the asset's underlying
            op. If set, Dagster will check that config provided for the op matches this schema and fail
            if it does not. If not set, Dagster will accept any config provided for the op.
        required_resource_keys (Optional[Set[str]]): Set of resource handles required by the underlying op.
        internal_asset_deps (Optional[Mapping[str, Set[AssetKey]]]): By default, it is assumed
            that all assets produced by a multi_asset depend on all assets that are consumed by that
            multi asset. If this default is not correct, you pass in a map of output names to a
            corrected set of AssetKeys that they depend on. Any AssetKeys in this list must be either
            used as input to the asset or produced within the op.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the assets.
        backfill_policy (Optional[BackfillPolicy]): The backfill policy for the op that computes the asset.
        op_tags (Optional[Dict[str, Any]]): A dictionary of tags for the op that computes the asset.
            Frameworks may expect and require certain metadata to be attached to a op. Values that
            are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.
        can_subset (bool): If this asset's computation can emit a subset of the asset
            keys based on the context.selected_asset_keys argument. Defaults to False.
        resource_defs (Optional[Mapping[str, object]]):
            (Beta) A mapping of resource keys to resources. These resources
            will be initialized during execution, and can be accessed from the
            context within the body of the function.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. This
            group name will be applied to all assets produced by this multi_asset.
        retry_policy (Optional[RetryPolicy]): The retry policy for the op that computes the asset.
        code_version (Optional[str]): Version of the code encapsulated by the multi-asset. If set,
            this is used as a default code version for all defined assets.
        specs (Optional[Sequence[AssetSpec]]): The specifications for the assets materialized
            by this function.
        check_specs (Optional[Sequence[AssetCheckSpec]]): Specs for asset checks that
            execute in the decorated function after materializing the assets.
        pool (Optional[str]): A string that identifies the concurrency pool that governs this
            multi-asset's execution.
        non_argument_deps (Optional[Union[Set[AssetKey], Set[str]]]): Deprecated, use deps instead.
            Set of asset keys that are upstream dependencies, but do not pass an input to the
            multi_asset.

    Examples:
        .. code-block:: python

            @multi_asset(
                specs=[
                    AssetSpec("asset1", deps=["asset0"]),
                    AssetSpec("asset2", deps=["asset0"]),
                ]
            )
            def my_function():
                asset0_value = load(path="asset0")
                asset1_result, asset2_result = do_some_transformation(asset0_value)
                write(asset1_result, path="asset1")
                write(asset2_result, path="asset2")

            # Or use IO managers to handle I/O:
            @multi_asset(
                outs={
                    "asset1": AssetOut(),
                    "asset2": AssetOut(),
                }
            )
            def my_function(asset0):
                asset1_value = do_some_transformation(asset0)
                asset2_value = do_some_other_transformation(asset0)
                return asset1_value, asset2_value
    """
    from dagster._core.execution.build_resources import wrap_resources_for_execution

    only_allow_hidden_params_in_kwargs(multi_asset, kwargs)

    args = DecoratorAssetsDefinitionBuilderArgs(
        name=name,
        op_description=description,
        specs=check.opt_sequence_param(specs, "specs", of_type=AssetSpec),
        check_specs_by_output_name=create_check_specs_by_output_name(check_specs),
        asset_out_map=check.opt_mapping_param(outs, "outs", key_type=str, value_type=AssetOut),
        upstream_asset_deps=_deps_and_non_argument_deps_to_asset_deps(
            deps=deps,
            non_argument_deps=_validate_hidden_non_argument_dep_param(
                kwargs.get("non_argument_deps")
            ),
        ),
        asset_deps=check.opt_mapping_param(
            internal_asset_deps, "internal_asset_deps", key_type=str, value_type=set
        ),
        asset_in_map=check.opt_mapping_param(ins, "ins", key_type=str, value_type=AssetIn),
        can_subset=can_subset,
        group_name=group_name,
        partitions_def=partitions_def,
        retry_policy=retry_policy,
        code_version=code_version,
        op_tags=op_tags,
        config_schema=check.opt_mapping_param(
            config_schema,  # type: ignore
            "config_schema",
            additional_message="Only dicts are supported for asset config_schema.",
        ),
        compute_kind=check.opt_str_param(kwargs.get("compute_kind"), "compute_kind"),
        required_resource_keys=check.opt_set_param(
            required_resource_keys, "required_resource_keys", of_type=str
        ),
        op_def_resource_defs=wrap_resources_for_execution(
            check.opt_mapping_param(resource_defs, "resource_defs", key_type=str)
        ),
        assets_def_resource_defs=wrap_resources_for_execution(
            check.opt_mapping_param(resource_defs, "resource_defs", key_type=str)
        ),
        backfill_policy=backfill_policy,
        decorator_name="@multi_asset",
        execution_type=AssetExecutionType.MATERIALIZATION,
        pool=pool,
    )

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        builder = DecoratorAssetsDefinitionBuilder.for_multi_asset(args=args, fn=fn)

        check.invariant(
            len(builder.overlapping_output_names) == 0,
            f"Check output names overlap with asset output names: {builder.overlapping_output_names}",
        )

        with disable_dagster_warnings():
            return builder.create_assets_definition()

    return inner


@overload
def graph_asset(
    compose_fn: Callable[..., Any],
) -> AssetsDefinition: ...


@overload
def graph_asset(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    config: Optional[Union[ConfigMapping, Mapping[str, Any]]] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    group_name: Optional[str] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    metadata: Optional[RawMetadataMapping] = ...,
    tags: Optional[Mapping[str, str]] = ...,
    owners: Optional[Sequence[str]] = None,
    kinds: Optional[AbstractSet[str]] = None,
    freshness_policy: Optional[FreshnessPolicy] = ...,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = ...,
    automation_condition: Optional[AutomationCondition] = ...,
    backfill_policy: Optional[BackfillPolicy] = ...,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = ...,
    check_specs: Optional[Sequence[AssetCheckSpec]] = None,
    code_version: Optional[str] = None,
    key: Optional[CoercibleToAssetKey] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]: ...


@hidden_param(
    param="freshness_policy",
    breaking_version="1.10.0",
    additional_warn_text="use freshness checks instead",
)
@hidden_param(
    param="auto_materialize_policy",
    breaking_version="1.10.0",
    additional_warn_text="use `automation_condition` instead",
)
def graph_asset(
    compose_fn: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    config: Optional[Union[ConfigMapping, Mapping[str, Any]]] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    group_name: Optional[str] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    metadata: Optional[RawMetadataMapping] = None,
    tags: Optional[Mapping[str, str]] = None,
    owners: Optional[Sequence[str]] = None,
    automation_condition: Optional[AutomationCondition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    check_specs: Optional[Sequence[AssetCheckSpec]] = None,
    code_version: Optional[str] = None,
    key: Optional[CoercibleToAssetKey] = None,
    kinds: Optional[AbstractSet[str]] = None,
    **kwargs,
) -> Union[AssetsDefinition, Callable[[Callable[..., Any]], AssetsDefinition]]:
    """Creates a software-defined asset that's computed using a graph of ops.

    This decorator is meant to decorate a function that composes a set of ops or graphs to define
    the dependencies between them.

    Args:
        name (Optional[str]): The name of the asset.  If not provided, defaults to the name of the
            decorated function. The asset's name must be a valid name in Dagster (ie only contains
            letters, numbers, and underscores) and may not contain Python reserved keywords.
        description (Optional[str]):
            A human-readable description of the asset.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        config (Optional[Union[ConfigMapping], Mapping[str, Any]):
            Describes how the graph underlying the asset is configured at runtime.

            If a :py:class:`ConfigMapping` object is provided, then the graph takes on the config
            schema of this object. The mapping will be applied at runtime to generate the config for
            the graph's constituent nodes.

            If a dictionary is provided, then it will be used as the default run config for the
            graph. This means it must conform to the config schema of the underlying nodes. Note
            that the values provided will be viewable and editable in the Dagster UI, so be careful
            with secrets.

            If no value is provided, then the config schema for the graph is the default (derived
            from the underlying nodes).
        key_prefix (Optional[Union[str, Sequence[str]]]): If provided, the asset's key is the
            concatenation of the key_prefix and the asset's name, which defaults to the name of
            the decorated function. Each item in key_prefix must be a valid name in Dagster (ie only
            contains letters, numbers, and underscores) and may not contain Python reserved keywords.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If
            not provided, the name "default" is used.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
        metadata (Optional[RawMetadataMapping]): Dictionary of metadata to be associated with
            the asset.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the asset. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        kinds (Optional[Set[str]]): A list of strings representing the kinds of the asset. These
            will be made visible in the Dagster UI.
        automation_condition (Optional[AutomationCondition]): The AutomationCondition to use
            for this asset.
        backfill_policy (Optional[BackfillPolicy]): The BackfillPolicy to use for this asset.
        code_version (Optional[str]): Version of the code that generates this asset. In
            general, versions should be set only for code that deterministically produces the same
            output when given the same inputs.
        key (Optional[CoeercibleToAssetKey]): The key for this asset. If provided, cannot specify key_prefix or name.

    Examples:
        .. code-block:: python

            @op
            def fetch_files_from_slack(context) -> pd.DataFrame:
                ...

            @op
            def store_files(files) -> None:
                files.to_sql(name="slack_files", con=create_db_connection())

            @graph_asset
            def slack_files_table():
                return store_files(fetch_files_from_slack())
    """
    only_allow_hidden_params_in_kwargs(graph_asset, kwargs)

    if compose_fn is None:
        return lambda fn: graph_asset(
            fn,  # type: ignore
            name=name,
            description=description,
            ins=ins,
            config=config,
            key_prefix=key_prefix,
            group_name=group_name,
            partitions_def=partitions_def,
            metadata=metadata,
            tags=tags,
            owners=owners,
            freshness_policy=kwargs.get("freshness_policy"),
            automation_condition=resolve_automation_condition(
                automation_condition, kwargs.get("auto_materialize_policy")
            ),
            backfill_policy=backfill_policy,
            resource_defs=resource_defs,
            check_specs=check_specs,
            code_version=code_version,
            key=key,
            kinds=kinds,
        )
    else:
        return graph_asset_no_defaults(
            compose_fn=compose_fn,
            name=name,
            description=description,
            ins=ins,
            config=config,
            key_prefix=key_prefix,
            group_name=group_name,
            partitions_def=partitions_def,
            metadata=metadata,
            tags=tags,
            owners=owners,
            freshness_policy=kwargs.get("freshness_policy"),
            automation_condition=resolve_automation_condition(
                automation_condition, kwargs.get("auto_materialize_policy")
            ),
            backfill_policy=backfill_policy,
            resource_defs=resource_defs,
            check_specs=check_specs,
            code_version=code_version,
            key=key,
            kinds=kinds,
        )


def graph_asset_no_defaults(
    *,
    compose_fn: Callable[..., Any],
    name: Optional[str],
    description: Optional[str],
    ins: Optional[Mapping[str, AssetIn]],
    config: Optional[Union[ConfigMapping, Mapping[str, Any]]],
    key_prefix: Optional[CoercibleToAssetKeyPrefix],
    group_name: Optional[str],
    partitions_def: Optional[PartitionsDefinition],
    metadata: Optional[RawMetadataMapping],
    tags: Optional[Mapping[str, str]],
    owners: Optional[Sequence[str]],
    freshness_policy: Optional[FreshnessPolicy],
    automation_condition: Optional[AutomationCondition],
    backfill_policy: Optional[BackfillPolicy],
    resource_defs: Optional[Mapping[str, ResourceDefinition]],
    check_specs: Optional[Sequence[AssetCheckSpec]],
    code_version: Optional[str],
    key: Optional[CoercibleToAssetKey],
    kinds: Optional[AbstractSet[str]],
) -> AssetsDefinition:
    ins = ins or {}
    named_ins = build_and_validate_named_ins(compose_fn, ins or {}, set())
    out_asset_key, _asset_name = resolve_asset_key_and_name_for_decorator(
        key=key,
        key_prefix=key_prefix,
        name=name,
        decorator_name="@graph_asset",
        fn=compose_fn,
    )

    keys_by_input_name = {input_name: asset_key for asset_key, (input_name, _) in named_ins.items()}
    partition_mappings = {
        input_name: asset_in.partition_mapping
        for input_name, asset_in in ins.items()
        if asset_in.partition_mapping
    }

    check_specs_by_output_name = validate_and_assign_output_names_to_check_specs(
        check_specs, [out_asset_key]
    )
    check_outs_by_output_name: Mapping[str, GraphOut] = {
        output_name: GraphOut() for output_name in check_specs_by_output_name.keys()
    }

    combined_outs_by_output_name: Mapping = {
        "result": GraphOut(),
        **check_outs_by_output_name,
    }

    validate_kind_tags(kinds)
    tags_with_kinds = {
        **(normalize_tags(tags, strict=True)),
        **{f"{KIND_PREFIX}{kind}": "" for kind in kinds or []},
    }

    op_graph = graph(
        name=out_asset_key.to_python_identifier(),
        description=description,
        config=config,
        ins={input_name: GraphIn() for _, (input_name, _) in named_ins.items()},
        out=combined_outs_by_output_name,
    )(compose_fn)
    return AssetsDefinition.from_graph(
        op_graph,
        keys_by_input_name=keys_by_input_name,
        keys_by_output_name={"result": out_asset_key},
        partitions_def=partitions_def,
        partition_mappings=partition_mappings if partition_mappings else None,
        group_name=group_name,
        metadata_by_output_name={"result": metadata} if metadata else None,
        tags_by_output_name={"result": tags_with_kinds} if tags_with_kinds else None,
        freshness_policies_by_output_name=(
            {"result": freshness_policy} if freshness_policy else None
        ),
        automation_conditions_by_output_name=(
            {"result": automation_condition} if automation_condition else None
        ),
        backfill_policy=backfill_policy,
        descriptions_by_output_name={"result": description} if description else None,
        resource_defs=resource_defs,
        check_specs=check_specs,
        owners_by_output_name={"result": owners} if owners else None,
        code_versions_by_output_name={"result": code_version} if code_version else None,
    )


def graph_multi_asset(
    *,
    outs: Mapping[str, AssetOut],
    name: Optional[str] = None,
    ins: Optional[Mapping[str, AssetIn]] = None,
    partitions_def: Optional[PartitionsDefinition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    group_name: Optional[str] = None,
    can_subset: bool = False,
    resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    check_specs: Optional[Sequence[AssetCheckSpec]] = None,
    config: Optional[Union[ConfigMapping, Mapping[str, Any]]] = None,
) -> Callable[[Callable[..., Any]], AssetsDefinition]:
    """Create a combined definition of multiple assets that are computed using the same graph of
    ops, and the same upstream assets.

    Each argument to the decorated function references an upstream asset that this asset depends on.
    The name of the argument designates the name of the upstream asset.

    Args:
        name (Optional[str]): The name of the graph.
        outs: (Optional[Dict[str, AssetOut]]): The AssetOuts representing the produced assets.
        ins (Optional[Mapping[str, AssetIn]]): A dictionary that maps input names to information
            about the input.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the assets.
        backfill_policy (Optional[BackfillPolicy]): The backfill policy for the asset.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. This
            group name will be applied to all assets produced by this multi_asset.
        can_subset (bool): Whether this asset's computation can emit a subset of the asset
            keys based on the context.selected_assets argument. Defaults to False.
        config (Optional[Union[ConfigMapping], Mapping[str, Any]):
            Describes how the graph underlying the asset is configured at runtime.

            If a :py:class:`ConfigMapping` object is provided, then the graph takes on the config
            schema of this object. The mapping will be applied at runtime to generate the config for
            the graph's constituent nodes.

            If a dictionary is provided, then it will be used as the default run config for the
            graph. This means it must conform to the config schema of the underlying nodes. Note
            that the values provided will be viewable and editable in the Dagster UI, so be careful
            with secrets.

            If no value is provided, then the config schema for the graph is the default (derived
                from the underlying nodes).
    """

    def inner(fn: Callable[..., Any]) -> AssetsDefinition:
        partition_mappings = {
            input_name: asset_in.partition_mapping
            for input_name, asset_in in (ins or {}).items()
            if asset_in.partition_mapping
        }

        named_ins = build_and_validate_named_ins(fn, ins or {}, set())
        keys_by_input_name = {
            input_name: asset_key for asset_key, (input_name, _) in named_ins.items()
        }
        named_outs = build_named_outs(outs)

        check_specs_by_output_name = validate_and_assign_output_names_to_check_specs(
            check_specs, list(named_outs.keys())
        )
        check_outs_by_output_name: Mapping[str, GraphOut] = {
            output_name: GraphOut() for output_name in check_specs_by_output_name.keys()
        }

        combined_outs_by_output_name = {
            **{output_name: GraphOut() for output_name, _ in named_outs.values()},
            **check_outs_by_output_name,
        }

        op_graph = graph(
            name=name or fn.__name__,
            out=combined_outs_by_output_name,
            config=config,
            ins={input_name: GraphIn() for _, (input_name, _) in named_ins.items()},
        )(fn)

        # source metadata from the AssetOuts (if any)
        metadata_by_output_name = {
            output_name: out.metadata
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.metadata is not None
        }

        # source freshness policies from the AssetOuts (if any)
        freshness_policies_by_output_name = {
            output_name: out.freshness_policy
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.freshness_policy is not None
        }

        # source auto materialize policies from the AssetOuts (if any)
        automation_conditions_by_output_name = {
            output_name: out.automation_condition
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.automation_condition is not None
        }

        # source descriptions from the AssetOuts (if any)
        descriptions_by_output_name = {
            output_name: out.description
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.description is not None
        }

        # source code versions from the AssetOuts (if any)
        code_versions_by_output_name = {
            output_name: out.code_version
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.code_version is not None
        }

        tags_by_output_name = {
            output_name: out.tags
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.tags is not None
        }

        owners_by_output_name = {
            output_name: out.owners
            for output_name, out in outs.items()
            if isinstance(out, AssetOut) and out.owners is not None
        }

        return AssetsDefinition.from_graph(
            op_graph,
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={
                output_name: asset_key for asset_key, (output_name, _) in named_outs.items()
            },
            partitions_def=partitions_def,
            partition_mappings=partition_mappings if partition_mappings else None,
            group_name=group_name,
            can_subset=can_subset,
            metadata_by_output_name=metadata_by_output_name,
            freshness_policies_by_output_name=freshness_policies_by_output_name,
            automation_conditions_by_output_name=automation_conditions_by_output_name,
            backfill_policy=backfill_policy,
            descriptions_by_output_name=descriptions_by_output_name,
            resource_defs=resource_defs,
            check_specs=check_specs,
            code_versions_by_output_name=code_versions_by_output_name,
            tags_by_output_name=tags_by_output_name,
            owners_by_output_name=owners_by_output_name,
        )

    return inner


def _deps_and_non_argument_deps_to_asset_deps(
    deps: Optional[Iterable[CoercibleToAssetDep]],
    non_argument_deps: Optional[Union[set[AssetKey], set[str]]],
) -> Optional[Iterable[AssetDep]]:
    """Helper function for managing deps and non_argument_deps while non_argument_deps is still an accepted parameter.
    Ensures only one of deps and non_argument_deps is provided, then converts the deps to AssetDeps.
    """
    if non_argument_deps is not None and deps is not None:
        raise DagsterInvalidDefinitionError(
            "Cannot specify both deps and non_argument_deps to @asset. Use only deps instead."
        )

    if deps is not None:
        return make_asset_deps(deps)

    if non_argument_deps is not None:
        check.set_param(non_argument_deps, "non_argument_deps", of_type=(AssetKey, str))
        return make_asset_deps(non_argument_deps)


def make_asset_deps(deps: Optional[Iterable[CoercibleToAssetDep]]) -> Optional[Iterable[AssetDep]]:
    if deps is None:
        return None

    return coerce_to_deps_and_check_duplicates(deps, key=None)
